"""Module for FastModelBuilder"""
import numpy as np
from scipy.stats import t, f

from organon.fl.mathematics.linearalgebra.blas_service_factory import BlasServiceFactory
from organon.fl.mathematics.sweep.covariance_generator import CovarianceGenerator
from organon.fl.mathematics.sweep.data_classes.linear_regression_algorithm_settings import \
    LinearRegressionAlgorithmSettings
from organon.fl.mathematics.sweep.data_classes.linear_regression_parameter import LinearRegressionParameter
from organon.fl.mathematics.sweep.data_classes.linear_regression_results import LinearRegressionResults
from organon.fl.mathematics.sweep.data_classes.linear_regression_step import LinearRegressionStep
from organon.fl.mathematics.sweep.data_classes.selected_attribute import SelectedAttribute
from organon.fl.mathematics.sweep.data_classes.transformation import Transformation
from organon.fl.mathematics.sweep.enums import AttributeSelectionStatus, TestStatisticsType, \
    RegressionAttributeSelectionMethod
from organon.fl.mathematics.sweep.matrices.i_matrix import IMatrix
from organon.fl.mathematics.sweep.sweep_operator import SweepOperator


class FastModelBuilder:
    """This is the class where everything gets together. It takes an input matrix. Calculates the attributes
    to be included or excluded every step according to settings. And of the method it returns the independent
    attributes that defines the target
    See ftp://public.dhe.ibm.com/software/analytics/spss/support/Stats/Docs/Statistics/Algorithms/13.0/regression.pdf
    for the more details."""

    def __init__(self, settings: LinearRegressionAlgorithmSettings, inputs: IMatrix, validation: IMatrix):
        """Initializes the model."""
        self._settings = settings
        self._inputs = inputs
        self._validation = validation
        self._target = self._settings.target
        self._weight = self._settings.weight

        bool(self._validation is not None)

        cols = self._inputs.shape[1]

        self._transformations = []
        for _col in range(cols):
            self._transformations.append(Transformation(0, 0, 0, False))

        blas_service = BlasServiceFactory.get_blas_service()

        self._training_sigma_generator = CovarianceGenerator(False, inputs, self._transformations, self._weight,
                                                             blas_service)
        self._training_sigma_generator.initialize()
        self._training_covariances = self._training_sigma_generator.covariance_info
        self._validation_covariances = None
        if self._validation is not None:
            self._validation_sigma_generator = CovarianceGenerator(False, validation, self._transformations,
                                                                   self._weight, blas_service)
            self._validation_sigma_generator.initialize()
            self._validation_covariances = self._validation_sigma_generator.covariance_info

        self._sweep_operator = SweepOperator(self._training_covariances.correlation_matrix,
                                             self._training_covariances.covariance_matrix, self._target, blas_service)

        self._regression_report = []
        self._regression_results = None

        self._swept = None
        self._non_swept = None

    def _validate(self):
        """Validates the inputs for the Model"""
        if self._settings is None:
            raise ValueError("Algorithm Settings can not be None as argument to Linear Regression object")

        if self._inputs is None:
            raise ValueError("Input matrix can not be None as argument to Linear Regression object")

        if self._target not in range(self._inputs.shape[1]):
            raise IndexError("Target is out of bounds for input matrix")

        if self._weight is None:
            raise ValueError("Weight can not be None as argument to Linear Regression object")

        if self._weight.shape[0] != self._inputs.shape[1]:
            raise ValueError("Weight shape and input shape don't fit. "
                             f"Weight should include {self._inputs.shape[1]} elements")

    def build_model(self):
        """Builds the model. Initializes the variables and calculates the diagonal entries of the covariance
        matrix and the target row and column. Starts the selection method according to settings. Then
        calculates the results. Finally sweep back the correlation matrix to initial matrix."""
        print("Building Model...")
        self._validate()
        self._training_sigma_generator.compute_variances()
        self._training_sigma_generator.compute_covariances(self._target)

        if self._validation is not None:
            self._validation_sigma_generator.compute_variances()
            self._validation_sigma_generator.compute_covariances(self._target)
        print("Variances and Covariances of target calculated")

        eligible_inputs = self._training_covariances.get_non_zero_variance_attributes()
        eligible_inputs.remove(self._target)
        self._swept = set()
        self._non_swept = set(eligible_inputs)

        if self._settings.attribute_selection_method == RegressionAttributeSelectionMethod.FORWARD:
            print("Forward Selection")
            self._forward_selection()

        if self._settings.attribute_selection_method == RegressionAttributeSelectionMethod.STEPWISE:
            print("Stepwise selection")
            self._stepwise_selection()
        print("Getting results...")
        self._regression_results = self._get_linear_regression_results()
        print("Sweeping back correlation matrix")
        self._sweep_fast_row(self._training_covariances.correlation_matrix, self._target, self._swept)

    @staticmethod
    def _tolerance_check(correlation_matrix: np.ndarray, swept: set, k: int, tolerance: float) -> bool:
        """Checks if the attribute can pass the tolerance check."""
        if correlation_matrix[k, k] < tolerance:
            return False

        for i in swept:
            if (correlation_matrix[i, i] - correlation_matrix[i, k] * correlation_matrix[k, i]) \
                    / correlation_matrix[k, k] * tolerance > 1.0:
                return False

        return True

    def _forward_selection(self):
        """Every time it gets the attribute. If there is no attribute that pass the conditions then selection
        is complete"""
        step = self._get_next_step()

        while True:
            print("Step = ", step)
            included = self._include_step()
            if included is None:
                break
            print("Including", included.attribute_name)
            self._update_reg_report(step, included)
            step += 1

    def _stepwise_selection(self):
        """Every step calculates the including and excluding attributes to the model. If there is not then exit.
        If they are the same we are in loop. Exit."""
        step = self._get_next_step()

        while True:
            print("Step = ", step)
            included = self._include_step()
            if included is not None:
                print("Including", included.attribute_name)
                self._update_reg_report(step, included)
                step += 1

            excluded = self._exclude_step()
            if excluded is not None:
                print("Excluding", excluded.attribute_name)
                self._update_reg_report(step, excluded)
                step += 1

            if included is None and excluded is None:
                break

            if excluded is not None and included is not None and excluded.attribute_name == included.attribute_name:
                break

    def _get_next_step(self) -> int:
        """Gets the step number according to the report."""
        return 1 if len(self._regression_report) == 0 else len(self._regression_report)

    def _update_reg_report(self, step: int, attribute: SelectedAttribute):
        """Updates the report with the new attribute"""
        anova = self._get_linear_regression_results()
        regression_step = LinearRegressionStep(step=step, selected_attribute=attribute, anova=anova)

        self._regression_report.append(regression_step)

    def _get_f_index_k_pairs(self, correlation_matrix, p_star):
        """Calculates the f_enter k pairs with f_enter values are eligible. Then returns the sorted list with
         highest f_enter value attribute first."""
        f_index_k_pairs = []
        for k in self._non_swept:
            # print("target = {}, k = {}".format(target, k))
            # print("calculating f_enter for ", k)
            v_k = correlation_matrix[self._target, k] * correlation_matrix[k, self._target] / correlation_matrix[k, k]
            delta = np.abs(correlation_matrix[self._target, self._target] - v_k)

            f_enter = float('+inf')
            # print("Delta for {0} = {1}".format(k, delta))

            if delta > np.finfo('float').eps:
                f_enter = (self._training_covariances.total_case_weight - p_star - 1) * v_k / delta

            if f_enter < self._settings.inclusion_f_statistics:
                continue

            base_condition = FastModelBuilder._tolerance_check(correlation_matrix, self._swept, k,
                                                               self._settings.tolerance)

            if correlation_matrix[self._target, k] > 0 and base_condition \
                    if self._settings.use_positive_coefficients else base_condition:
                f_index_k_pairs.append((f_enter, k))

        return sorted(f_index_k_pairs, key=lambda x: x[0], reverse=True)

    def _get_included_attribute(self, correlation_matrix, f_index_k_pairs, pre_r2_training, pre_r2_validation):
        """For the values that are eligible it returns the first element that changes the performance large
        enough according to sweep operator"""
        for pair in f_index_k_pairs:
            f_statistics = pair[0]
            k = pair[1]
            print(f"Got {k} with {f_statistics}")

            post_r2_training = self._get_training_r_squared(correlation_matrix, self._target, k)
            r2_delta_training = post_r2_training - pre_r2_training

            if r2_delta_training < self._settings.min_r_squared_change_in_train_set:
                break

            max_vif = FastModelBuilder._get_max_v_if(FastModelBuilder.updated_diagonals(correlation_matrix,
                                                                                        self._swept, k))

            if max_vif > self._settings.maximum_vif:
                continue

            if self._settings.use_positive_coefficients:
                updated_target_row = FastModelBuilder._updated_target_row(correlation_matrix, self._swept, self._target,
                                                                          k)
                if not FastModelBuilder._all_positive(updated_target_row):
                    continue
            self._training_sigma_generator.compute_covariances(k)
            if self._validation is not None:
                self._validation_sigma_generator.compute_covariances(k)
            # print("before sweep for ", k)
            # print(self._training_covariances.correlation_matrix)
            self._sweep_operator.sweep(k)
            # print("after sweep for ", k)
            # print(self._training_covariances.correlation_matrix)
            self._include(k)

            performance_change_is_large = True
            post_r2_validation = float('nan')
            r2_delta_validation = float('nan')

            if self._validation is not None:
                post_r2_validation = self._get_validation_r_squared()
                r2_delta_validation = post_r2_validation - pre_r2_validation

                if r2_delta_validation < self._settings.min_r_squared_change_in_validation_set:
                    performance_change_is_large = False

            if performance_change_is_large:
                print(f"performance change for {k} is okay")
                return SelectedAttribute(
                    attribute_name=k,
                    selection_status=AttributeSelectionStatus.INCLUDED,
                    selection_test_statistics=TestStatisticsType.F_STATISTICS,
                    selection_statistics=f_statistics,
                    updated_training_performance=post_r2_training,
                    updated_validation_performance=post_r2_validation if self._validation is not None else float('nan'),
                    performance_change_in_training_set=r2_delta_training,
                    performance_change_in_validation_set
                    =r2_delta_validation if self._validation is not None else float('nan')
                )

            # print("Sweep back for", k)
            self._sweep_operator.sweep(k)
            self._exclude(k)
        return None

    def _include_step(self) -> SelectedAttribute:
        """It selects the included attribute according to f_enter,k pairs and the performance change with sweep
        operator."""
        print("Checking any variable to be included")
        p_star = 1 + len(self._swept) if self._settings.use_intercept else len(self._swept)

        correlation_matrix = self._training_covariances.correlation_matrix

        pre_r2_training = 0.0
        pre_r2_validation = 0.0

        if len(self._swept) > 0:
            pre_r2_training = self._get_training_r_squared(correlation_matrix, self._target)

            if self._validation is not None:
                pre_r2_validation = self._get_validation_r_squared()

        f_index_k_pairs = self._get_f_index_k_pairs(correlation_matrix, p_star)

        return self._get_included_attribute(correlation_matrix, f_index_k_pairs, pre_r2_training, pre_r2_validation)

    def _exclude_step(self) -> SelectedAttribute:
        """Does the same thing with the include step according to f_remove values."""
        print("Checking any variable to be excluded")

        p_star = 1 + len(self._swept) if self._settings.use_intercept else len(self._swept)

        correlation_matrix = self._training_covariances.correlation_matrix

        f_out = self._settings.exclusion_f_statistics

        f_index_k_pairs = []

        for k in self._swept:

            v_k = np.abs(correlation_matrix[self._target, k] * correlation_matrix[k, self._target]
                         / correlation_matrix[k, k])

            f_remove = ((self._training_covariances.total_case_weight - p_star) * v_k
                        / correlation_matrix[self._target, self._target])

            if f_remove < f_out:
                f_index_k_pairs.append((f_remove, k))

        if len(f_index_k_pairs) <= 0:
            print("Nothing excluded")
            return None

        min_pair = min(f_index_k_pairs, key=lambda x: x[0])

        selected_index = min_pair[1]
        f_out_statistics = min_pair[0]

        print(f"Got {selected_index} with {f_out_statistics}")

        pre_r2_training = 0.0
        pre_r2_validation = 0.0

        if len(self._swept) > 0:
            pre_r2_training = self._get_training_r_squared(correlation_matrix, self._target)

            if self._validation is not None:
                pre_r2_validation = self._get_validation_r_squared()

        self._sweep_operator.sweep(selected_index)
        self._exclude(selected_index)

        post_r2_training = self._get_training_r_squared(correlation_matrix, self._target)
        post_r2_validation = self._get_validation_r_squared()

        return SelectedAttribute(attribute_name=selected_index,
                                 selection_status=AttributeSelectionStatus.EXLUDED,
                                 selection_test_statistics=TestStatisticsType.F_STATISTICS,
                                 selection_statistics=f_out_statistics,
                                 updated_training_performance=post_r2_training,
                                 updated_validation_performance
                                 =post_r2_validation if self._validation is not None else float(
                                     'nan'),
                                 performance_change_in_training_set=post_r2_training - pre_r2_training,
                                 performance_change_in_validation_set
                                 =post_r2_validation - pre_r2_validation if self._validation is not None else float(
                                     'nan'))

    @staticmethod
    def _get_training_r_squared(correlation_matrix: np.ndarray, target: int, *arg) -> float:
        """Returns the training r squared. See the paper."""
        if len(arg) == 0:
            swept_r = correlation_matrix
            return 1.0 - swept_r[target, target]

        if len(arg) == 1:
            non_swept_r = correlation_matrix
            k = arg[0]

            return 1.0 - non_swept_r[target, target] + non_swept_r[target, k] * non_swept_r[k, target] / non_swept_r[
                k, k]
        return None

    def _include(self, attribute: int):
        """Includes the attribute"""
        self._swept.add(attribute)
        self._non_swept.remove(attribute)

    def _exclude(self, attribute: int):
        """Excludes the attribute"""
        self._swept.remove(attribute)
        self._non_swept.add(attribute)

    def _get_validation_r_squared(self) -> float:
        """Returns the validation r squared"""
        model = self._get_linear_regression_results()

        linear_form = {x: model.parameters[x].estimate for x in model.parameters}

        covariance = self._validation_covariances.get_covariances(linear_form, self._target)
        target_variance = self._validation_covariances.covariance_matrix[self._target, self._target]
        linear_form_variance = self._validation_covariances.get_variances(linear_form)

        target_variance_is_non_zero = ~np.isinf(target_variance) and ~np.isnan(target_variance) and target_variance > 0
        linear_form_variance_is_non_zero = (~np.isinf(linear_form_variance) and ~np.isnan(linear_form_variance) and
                                            linear_form_variance > 0)
        covariance_is_not_extreme = ~np.isinf(covariance) and ~np.isnan(covariance)

        if target_variance_is_non_zero and linear_form_variance_is_non_zero and covariance_is_not_extreme:
            rho = covariance / np.sqrt(linear_form_variance * target_variance)
            rho_is_not_extreme = ~np.isinf(rho) and ~np.isnan(rho)
            return rho * rho if rho_is_not_extreme else 0.0

        return 0.0

    @staticmethod
    def _get_max_v_if(updated_diagonals: dict) -> float:
        """Returns the max of the updated diagonals"""
        return float('-inf') if len(updated_diagonals) < 2 else max(updated_diagonals.values())

    @staticmethod
    def updated_diagonals(correlation_matrix: np.ndarray, swept: set, k: int) -> dict:
        """Calculates the updated diagonal values for the correlation matrix"""
        diags = {}
        for i in swept:
            updated_r = correlation_matrix[i, i] - correlation_matrix[i, k] * correlation_matrix[k, i] / \
                        correlation_matrix[k, k]
            diags[i] = updated_r
        diags[k] = 1.0 / correlation_matrix[k, k]

        return diags

    @staticmethod
    def _all_positive(swept_diagonals: dict) -> bool:
        """Check if the all values are positive"""
        return all(value > 0 for value in swept_diagonals.values())

    @staticmethod
    def _updated_target_row(correlation_matrix: np.ndarray, swept: set, target: int, k: int) -> dict:
        """Calculates the updated row for the correlation matrix"""
        diags = {}
        for i in swept:
            updated_r = correlation_matrix[i, i] - correlation_matrix[i, k] * correlation_matrix[k, i] / \
                        correlation_matrix[k, k]
            diags[i] = updated_r
        diags[k] = correlation_matrix[target, k] / correlation_matrix[k, k]

        return diags

    def _get_linear_regression_parameters(self):
        """Calculates the statistical info for the every swept attribute."""
        covariance_matrix = self._training_covariances.covariance_matrix
        correlation_matrix = self._training_covariances.correlation_matrix
        total_case_weight = self._training_covariances.total_case_weight
        p_star = len(self._swept) + 1 if self._settings.use_intercept else len(self._swept)
        parameters = {}
        t_dist = t(total_case_weight - p_star)

        for i in self._swept:
            estimate = correlation_matrix[self._target, i] * np.sqrt(
                covariance_matrix[self._target, self._target] / covariance_matrix[i, i])
            standard_error = np.sqrt(
                correlation_matrix[i, i] * correlation_matrix[self._target, self._target]
                * covariance_matrix[self._target, self._target]) \
                             / (covariance_matrix[i, i] * (total_case_weight - p_star))
            sigma = 0.0 if standard_error <= 0 or np.isnan(standard_error) else standard_error
            t_statistics = float('+inf') if sigma <= 0 else estimate / sigma
            absolute_t_statistics = np.abs(t_statistics)
            vif = correlation_matrix[i, i]
            p_value = 0.0 if np.isinf(absolute_t_statistics) else 1 - t_dist.cdf(absolute_t_statistics)
            parameters[i] = LinearRegressionParameter(estimate, standard_error, t_statistics, p_value, vif,
                                                      float('nan'),
                                                      float('nan'),
                                                      float('nan'))
        return parameters

    def _get_linear_regression_results(self) -> LinearRegressionResults:
        """Calculates the statistical info for the results."""
        covariance_matrix = self._training_covariances.covariance_matrix
        correlation_matrix = self._training_covariances.correlation_matrix
        total_case_weight = self._training_covariances.total_case_weight
        p_star = len(self._swept) + 1 if self._settings.use_intercept else len(self._swept)

        model_detail = LinearRegressionResults()
        model_detail.sample_size = total_case_weight
        model_detail.model_degrees_of_freedom = len(self._swept)
        model_detail.error_degrees_of_freedom = total_case_weight - p_star
        model_detail.r_square = 1 - correlation_matrix[self._target, self._target]
        model_detail.adjusted_r_square = model_detail.r_square - (len(self._swept) * (1.0 - model_detail.r_square)) / (
                total_case_weight - p_star)
        model_detail.residual_sum_of_squares = correlation_matrix[self._target, self._target] \
                                               * (total_case_weight - 1) * covariance_matrix[self._target, self._target]

        model_detail.explained_sum_of_squares = model_detail.r_square * (total_case_weight - 1) * covariance_matrix[
            self._target, self._target]
        model_detail.total_sum_of_squares = (model_detail.residual_sum_of_squares +
                                             model_detail.explained_sum_of_squares)
        model_detail.mean_square_regression = (model_detail.explained_sum_of_squares /
                                               model_detail.model_degrees_of_freedom)
        model_detail.mean_square_error = model_detail.residual_sum_of_squares / model_detail.error_degrees_of_freedom

        if model_detail.model_degrees_of_freedom > 0:
            f_distribution = f(model_detail.model_degrees_of_freedom, model_detail.error_degrees_of_freedom)

            f_statistics = model_detail.mean_square_regression / model_detail.mean_square_error \
                if model_detail.mean_square_error > 0 else float('+inf')
            p_value = 0.0 if np.isinf(f_statistics) else 1.0 - f_distribution.cdf(f_statistics)
            model_detail.f_statistics = f_statistics
            model_detail.f_statistics_p_value = p_value

        model_detail.parameters = self._get_linear_regression_parameters()

        means = self._training_covariances.means
        if self._settings.use_intercept:
            bias = means[self._target]

            for i in model_detail.parameters:
                attribute_mean = means[i]
                bias -= model_detail.parameters[i].estimate * attribute_mean

            model_detail.bias_dict['estimate'] = bias

        else:
            model_detail.bias_dict['estimate'] = 0.0

        model_detail.bias_dict['p_value'] = 0.0
        model_detail.bias_dict['standard_error'] = 0.0
        model_detail.bias_dict['t_statistics'] = 0.0

        return model_detail

    def _sweep_fast_row(self, matrix: np.ndarray, target: int, swept_list: set):
        """Calls sweep algorithm for the sweeping back all the swept attributes."""
        for row in swept_list:
            self._sweep_operator.sweep_fast_row(matrix, row, target, swept_list, matrix.shape[1])



"""Includes DqComparisonParameters class."""
from typing import TypeVar, Generic, List

from organon.idq.domain.businessobjects.dq_test_group import DqTestGroup
from organon.idq.domain.settings.abstractions.dq_base_input_source_settings import DqBaseInputSourceSettings
from organon.idq.domain.settings.dq_comparison_column_info import DqComparisonColumnInfo

T = TypeVar("T", bound=DqBaseInputSourceSettings)


class DqBaseComparisonParameters(Generic[T]):
    """Dq comparison settings"""

    def __init__(self):
        self.comparison_columns: List[DqComparisonColumnInfo] = None
        self.maximum_nom_cardinality: int = None
        self.minimum_cardinality: int = None
        self.traffic_light_threshold_yellow: float = None
        self.traffic_light_threshold_green: float = None
        self.psi_threshold_yellow: float = None
        self.psi_threshold_green: float = None
        self.z_score: float = None
        self.test_groups: List[DqTestGroup] = None



"""
This module includes TransactionFileRecordReader class.
"""
from datetime import datetime
from typing import Dict, Callable, Tuple, List, Optional

import numpy as np
import pandas as pd

from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.afe.dataaccess.helpers.record_source_helper import validate_entity_column
from organon.afe.dataaccess.services.base_record_reader import BaseRecordReader
from organon.afe.domain.common import afe_date_helper
from organon.afe.domain.enums.record_source_type import RecordSourceType
from organon.afe.domain.modelling.businessobjects.transaction_file_record_collection import \
    TransactionFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_stats import TransactionFileStats
from organon.afe.domain.settings.afe_date_column import AfeDateColumn
from organon.afe.domain.settings.afe_reading_settings import AfeDataReadingSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.fl.core.businessobjects.date_interval import DateInterval
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import date_helper


class TransactionFileRecordReader(BaseRecordReader):
    # pylint: disable=too-many-instance-attributes
    """
    Reads a transaction file source and creates a TransactionRecordCollection
    """
    UNKNOWN_DIMENSION_VALUE = -2

    def __init__(self,
                 trx_record_source: RecordSource,
                 date_columns: Dict[str, AfeDateColumn],
                 trx_entity_column: str,
                 dimension_columns: List[str] = None,
                 quantity_columns: List[str] = None,
                 **kwargs):
        super().__init__(trx_record_source)
        self._date_columns: Dict[str, AfeDateColumn] = date_columns
        self._trx_entity_column: str = trx_entity_column
        self._trx_record_source = trx_record_source
        self._distinct_entity_source: RecordSource = kwargs.get("distinct_entity_source")
        self._entity_source_entity_column: str = kwargs.get("entity_source_entity_column")
        self._date_col_map: Dict[str, int] = self._create_map(list(date_columns.keys()))

        self._interval_per_entity_per_date_column: Dict[str, Dict[str, DateInterval]] = \
            kwargs.get("interval_per_entity_per_date_column")

        self._interval_per_date_column: Dict[str, Tuple[datetime, datetime]] = kwargs.get("interval_per_date_column")

        if self._interval_per_date_column is not None and self._interval_per_entity_per_date_column is not None:
            raise ValueError("Please give either interval_per_date_column or interval_per_entity_per_date_column")

        self._min_max_dates_per_date_column: Dict[str, Tuple[datetime, datetime]] = \
            kwargs.get("min_max_dates_per_date_column")

        self._max_number_of_transaction_samples: int = kwargs.get("max_number_of_transaction_samples",
                                                                  np.iinfo(np.int64).max)

        self._reading_settings: AfeDataReadingSettings = kwargs.get("reading_settings",
                                                                    AfeDataReadingSettings(
                                                                        number_of_rows_per_step=100000))

        self._dimension_compression_ratio: float = kwargs.get("dimension_compression_ratio", 1.0)
        self._dimension_max_cardinality: int = kwargs.get("dimension_max_cardinality", 200)

        self._trx_file_stats: TransactionFileStats = kwargs.get("transaction_file_stats")

        self._func_per_date_column = TransactionFileRecordReader.__get_func_per_date_col_map(date_columns)

        if dimension_columns:
            self._d_map: Dict[str, int] = self._create_map(dimension_columns)
            self._non_empty_d_cols: List[str] = [elem for elem in self._d_map if
                                                 elem != AfeStaticObjects.empty_dimension_column]
        if quantity_columns:
            self._q_map: Dict[str, int] = self._create_map(quantity_columns)
            self._non_empty_q_cols: List[str] = [elem for elem in self._q_map if
                                                 elem != AfeStaticObjects.empty_quantity_column]

    @property
    def trx_record_source(self) -> RecordSource:
        """Return value of private attribute '__trx_record_source'"""
        return self._trx_record_source

    @property
    def distinct_entity_source(self) -> RecordSource:
        """Return value of private attribute '__distinct_entity_source'"""
        return self._distinct_entity_source

    @property
    def entity_source_entity_column(self) -> str:
        """Return value of private attribute '__entity_source_entity_column'"""
        return self._entity_source_entity_column

    @property
    def interval_per_entity_per_date_column(self) -> Dict[str, Dict[str, DateInterval]]:
        """Return value of private attribute '__interval_per_entity'"""
        return self._interval_per_entity_per_date_column

    @property
    def date_col_map(self) -> Dict[str, int]:
        """Return value of private attribute '__date_col_map'"""
        return self._date_col_map

    @property
    def d_map(self) -> Dict[str, int]:
        """Return value of private attribute '__d_map'"""
        return self._d_map

    @property
    def q_map(self) -> Dict[str, int]:
        """Return value of private attribute '__q_map'"""
        return self._q_map

    @property
    def func_per_date_column(self) -> Dict[str, Callable[[str], datetime]]:
        """Return value of private attribute '__func'"""
        return self._func_per_date_column

    @staticmethod
    def __get_func_per_date_col_map(date_columns: Dict[str, AfeDateColumn]):
        func_per_date_column = {}
        for date_column in date_columns.values():
            func = afe_date_helper.get_str_to_date_converter(date_column.date_column_type,
                                                             date_column.custom_format)
            func_per_date_column[date_column.column_name] = func
        return func_per_date_column

    def __get_date_interval_for_all_entities(self) -> Dict[str, Tuple[datetime, datetime]]:
        """
        Finds overall minimum and maximum dates from date boundaries of all entities
        :return: tuple containing min_date and max_date
        """
        date_interval_per_date_column = {}

        for date_column in self._date_columns.values():
            if date_column is not None:
                date_col_name = date_column.column_name
                min_date_ts: int = np.iinfo(np.int64).max
                max_date_ts: int = np.iinfo(np.int64).min
                # pylint: disable=unsubscriptable-object
                interval_per_entity = self._interval_per_entity_per_date_column[date_col_name]
                for item in interval_per_entity:
                    target_interval: DateInterval = interval_per_entity[item]
                    if target_interval.lower_bound < min_date_ts:
                        min_date_ts = target_interval.lower_bound
                    if target_interval.upper_bound > max_date_ts:
                        max_date_ts = target_interval.upper_bound
                min_date = date_helper.get_integer_as_date(min_date_ts)
                max_date = date_helper.get_integer_as_date(max_date_ts)
                date_interval_per_date_column[date_col_name] = (min_date, max_date)
        return date_interval_per_date_column

    def read_db_source(self) -> TransactionFileRecordCollection:
        """
        Reads database source and creates a transaction record collection.
        :return: transaction record collection
        """
        raise NotImplementedError

    def read_dataframe_source(self):
        """
        Reads pandas DataFrame source.
        :return: TransactionFileRecordCollection containing DataFrame records
        """
        if self._distinct_entity_source is not None and \
                self._distinct_entity_source.get_type() != RecordSourceType.DATA_FRAME:
            raise KnownException("Distinct entity source should be a dataframe")
        data_frame: pd.DataFrame = self._trx_record_source.source
        collection = self._initialize_collection(len(data_frame))
        unique_entities_set = None
        if self.distinct_entity_source is not None:
            unique_entities_set = self._get_unique_entities_set()
        self.__get_records_from_dataframe(data_frame, collection,
                                          self._max_number_of_transaction_samples,
                                          unique_entities_set)

        if self._trx_file_stats is None:
            self._collection_stats_build_indices(collection)
        collection.convert_entity_index_lists_to_array()
        return collection

    def read_text_source(self):
        """
        Reads text source.
        :return: TransactionFileRecordCollection containing text records
        """
        if self._distinct_entity_source is not None and \
                self._distinct_entity_source.get_type() != RecordSourceType.DATA_FRAME:
            raise KnownException("Distinct entity source should be a dataframe")

        created_records_num = 0
        chunksize = self._reading_settings.number_of_rows_per_step
        record_source = self._trx_record_source
        max_trx_samples = self._max_number_of_transaction_samples

        with open(record_source.source, "r", encoding="utf8") as file:
            row_count = sum(1 for row in file if row.strip() != "") - 1  # subtract for header line

        record_count = min(max_trx_samples, row_count)
        collection = self._initialize_collection(record_count)
        unique_entities_set: set = None
        if self.distinct_entity_source is not None:
            unique_entities_set = self._get_unique_entities_set()
        with open(record_source.source, encoding="utf-8") as source_file:
            for data_frame in pd.read_csv(source_file, sep=",", chunksize=chunksize):
                records_needed = max_trx_samples - created_records_num
                chunk_records_num = self.__get_records_from_dataframe(data_frame, collection,
                                                                      min(records_needed, chunksize),
                                                                      unique_entities_set)
                created_records_num += chunk_records_num
                if created_records_num >= max_trx_samples:
                    break

        if self._trx_file_stats is None:
            self._collection_stats_build_indices(collection)
        collection.convert_entity_index_lists_to_array()
        return collection

    def __get_records_from_dataframe(self, data_frame: pd.DataFrame, collection: TransactionFileRecordCollection,
                                     max_record_num: int,
                                     unique_entities_set: set = None):
        """
        Inserts records into given collection by reading the DataFrame

        :param set unique_entities_set: Set of unique entities to be read
        :param max_record_num: Maximum number of records to be created
        :param pd.DataFrame data_frame: DataFrame to be read
        :param TransactionFileRecordCollection collection: Collection to store records
        """
        trx_entity_column: str = self._trx_entity_column
        validate_entity_column("transaction", data_frame, trx_entity_column)
        trx_date_columns = [date_column.column_name for date_column in self._date_columns.values()]

        columns_ordered = [trx_entity_column]
        if len(trx_date_columns) > 0:
            columns_ordered.extend(trx_date_columns)

        columns_ordered.extend(self._non_empty_q_cols)
        columns_ordered.extend(self._non_empty_d_cols)
        column_index_map = {column: columns_ordered.index(column) for column in columns_ordered}
        created_records_num = 0

        for row_iter in data_frame.iterrows():
            frame_row = row_iter[1]
            entity_id = str(TransactionFileRecordReader.__get_col_value_from_row(frame_row, trx_entity_column))
            if unique_entities_set is not None and entity_id not in unique_entities_set:
                continue
            row = [entity_id]

            add_to_collection = True
            for date_col in trx_date_columns:
                date_value = self.func_per_date_column[date_col](str(
                    TransactionFileRecordReader.__get_col_value_from_row(frame_row, date_col)))
                row.append(date_value)
                if not self.__check_add_to_collection(date_col, date_value):
                    add_to_collection = False
                    break

            if add_to_collection:
                self.__append_column_ordered_d_and_q_column(row, frame_row)
                record_created = self.get_record_from_row(collection, row, column_index_map, convert_dates=False)
                if record_created:
                    created_records_num += 1
                    if created_records_num >= max_record_num:
                        break
        return created_records_num

    @staticmethod
    def __get_col_value_from_row(row, column_name: str):
        try:
            return row[column_name]
        except KeyError as key_err:
            raise KnownException(f"Column {key_err.args[0]} could not be found in transaction data source") from key_err

    def __check_add_to_collection(self, date_col: str, date_val: datetime):
        """Checks if row should be add to record collection"""
        min_max_dates_per_date_column = self._min_max_dates_per_date_column
        if min_max_dates_per_date_column is None or date_col not in min_max_dates_per_date_column:
            return True
        min_date, max_date = min_max_dates_per_date_column[date_col]
        if min_date is not None and date_val < min_date:
            return False
        if max_date is not None and date_val > max_date:
            return False

        return True

    def __append_column_ordered_d_and_q_column(self, row, frame_row):
        for col in self._non_empty_q_cols:
            row.append(str(TransactionFileRecordReader.__get_col_value_from_row(frame_row, col)))
        for col in self._non_empty_d_cols:
            col_value = TransactionFileRecordReader.__get_col_value_from_row(frame_row, col)
            if col_value is None:
                row.append(col_value)
            else:
                row.append(str(col_value))

    def get_record_from_row(self, collection: TransactionFileRecordCollection, row, column_positions: Dict[str, int],
                            convert_dates: bool = True):
        """
        Creates a TransactionFileRecord given a row of values
        :param convert_dates: Should be given false if date column values are already datetime objects
        :param collection: Collection instance to be filled
        :param row: A subscriptable storing entity id, date , quantity values, dimension values
        :param column_positions: Dictionary for column_name-row_index mapping
        :return: True if record was created successfully
        """
        entity_id = str(row[0])
        add_to_collection, trx_dates = self.__check_and_get_date_values(row, entity_id, column_positions, convert_dates)

        if not add_to_collection:
            return False

        q_array: List[None or int] = [None] * len(self._q_map)
        default_q_index = self._q_map[AfeStaticObjects.empty_quantity_column]
        q_array[default_q_index] = 1.0
        self.__set_q_array(column_positions, q_array, row)

        d_array = [0] * len(self._d_map)
        default_d_column: str = AfeStaticObjects.empty_dimension_column

        default_d_index = self._d_map[AfeStaticObjects.empty_dimension_column]
        d_array[default_d_index] = 0

        self.__set_d_array(column_positions, d_array, row, collection, default_d_column)

        collection.append(entity_id, trx_dates, q_array, d_array)
        return True

    def __set_q_array(self, column_positions, q_array, row):
        for q_column in self._non_empty_q_cols:
            pos: int = column_positions[q_column]
            value = row[pos]
            if value is None:
                continue
            index: int = self._q_map[q_column]
            q_array[index] = value

    def __set_d_array(self, column_positions, d_array, row, collection, default_d_column):
        if self._trx_file_stats is None:
            collection.transaction_file_stats.increment(default_d_column, "")
            for d_column in self._non_empty_d_cols:
                pos = column_positions[d_column]
                index: int = self._d_map[d_column]
                d_val = row[pos]
                if d_val is not None:
                    d_val: str = str(row[pos])
                collection.transaction_file_stats.increment(d_column, d_val)
                d_array[index] = collection.transaction_file_stats.get_index(d_column, d_val)

        else:
            for d_column in self._non_empty_d_cols:
                pos = column_positions[d_column]
                index: int = self._d_map[d_column]
                d_val = row[pos]
                if d_val is not None:
                    d_val: str = str(row[pos])
                try:
                    d_array[index] = self._trx_file_stats.get_index(d_column, d_val)
                except KeyError:
                    d_array[index] = self.UNKNOWN_DIMENSION_VALUE

    def __check_and_get_date_values(self, row, entity_id, column_positions, convert_dates):
        if len(self._date_columns) == 0:
            return True, []
        trx_dates: List[None or int] = [None] * len(self._date_col_map)
        date_cond = False

        for date_column in self._date_columns.values():
            pos: int = column_positions[date_column.column_name]
            date_val = self._func_per_date_column[date_column.column_name](str(row[pos])) \
                if convert_dates else row[pos]
            value = date_helper.get_date_as_integer(date_val)
            if value is None:
                continue
            index: int = self._date_col_map[date_column.column_name]

            if self._interval_per_date_column is not None:
                interval = self._interval_per_date_column[date_column.column_name]
                lower, upper = interval[0], interval[1]
                lower_bound_ok = True
                if lower is not None and date_val < lower:
                    lower_bound_ok = False
                upper_bound_ok = True
                if upper is not None and date_val > upper:
                    upper_bound_ok = False
                if lower_bound_ok and upper_bound_ok:
                    date_cond = True

            elif self._interval_per_entity_per_date_column is not None:
                # pylint: disable=unsubscriptable-object
                interval_per_entity = self._interval_per_entity_per_date_column[date_column.column_name]
                if entity_id in interval_per_entity:
                    if not date_cond and interval_per_entity[entity_id].contains(value):
                        date_cond = True
            else:
                date_cond = True
            trx_dates[index] = value

        return date_cond, trx_dates

    def _initialize_collection(self, record_count: int):
        """Creates and initializes a TransactionFileRecordCollection"""
        collection = TransactionFileRecordCollection(self._date_col_map,
                                                     self.d_map,
                                                     self.q_map,
                                                     record_count)
        default_d_column: str = AfeStaticObjects.empty_dimension_column
        if self._trx_file_stats is None:
            stats: TransactionFileStats = TransactionFileStats()
            stats.add(default_d_column)
            for dimension_column in self._non_empty_d_cols:
                stats.add(dimension_column)
            collection.transaction_file_stats = stats
        return collection

    def _get_intervals_per_date_column(self) -> Optional[Dict[str, Tuple[datetime, datetime]]]:
        """Returns minimum and maximum dates of concern for transaction"""
        if self._interval_per_date_column is None and self._interval_per_entity_per_date_column is None:
            return {}

        if self._interval_per_date_column is not None:
            interval_per_date_col = self._interval_per_date_column
        else:
            interval_per_date_col = self.__get_date_interval_for_all_entities()

        min_max_dates_per_date_column = {}

        for date_column in self._date_columns.values():
            if date_column is not None:
                date_col_name = date_column.column_name
                date_interval = interval_per_date_col[date_col_name]
                max_date: datetime = date_interval[1]
                min_date: datetime = date_interval[0]

                # pylint: disable=unsupported-membership-test,unsubscriptable-object
                if self._min_max_dates_per_date_column is not None and \
                        date_col_name in self._min_max_dates_per_date_column:
                    exp_min_date, exp_max_date = self._min_max_dates_per_date_column[date_col_name]

                    if exp_max_date is not None and exp_max_date < max_date:
                        max_date = exp_max_date
                    if exp_min_date is not None and min_date < exp_min_date:
                        min_date = exp_min_date

                min_max_dates_per_date_column[date_col_name] = (min_date, max_date)
        return min_max_dates_per_date_column

    def _collection_stats_build_indices(self, collection: TransactionFileRecordCollection):
        """builds indices for a trx file"""
        if collection.actual_record_count == 0:
            raise KnownException("There are no eligible records in collection")
        collection.transaction_file_stats.build_indices(
            self._dimension_compression_ratio,
            self._dimension_max_cardinality)

    def _get_unique_entities_set(self):
        return set(self._distinct_entity_source.source[self._entity_source_entity_column])

    @staticmethod
    def _create_map(column_list: List[str]) -> Dict[str, int]:
        _dict = {}
        index = 0
        for column in column_list:
            _dict[column] = index
            index += 1
        return _dict



"""Includes Modeller enum class."""
from enum import Enum, auto

from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType


class Modeller(Enum):
    """Modellers"""
    LOGISTIC_REGRESSION_CLASSIFIER = auto()
    GAM_CLASSIFIER = auto()
    RF_CLASSIFIER = auto()
    GBM_CLASSIFIER = auto()
    LIGHTGBM_CLASSIFIER = auto()
    XGBOOST_CLASSIFIER = auto()
    MULTI_LAYER_PERCEPTRON_CLASSIFIER = auto()
    RIDGE = auto()
    GAM_REGRESSOR = auto()
    LASSO = auto()
    RANDOM_FOREST_REGRESSOR = auto()
    GBM_REGRESSOR = auto()
    LIGHTGBM_REGRESSOR = auto()
    XGBOOST_REGRESSOR = auto()
    MULTI_LAYER_PERCEPTRON_REGRESSOR = auto()

    def get_modeller_type(self):
        """Returns type of modeller"""
        if self in [Modeller.LOGISTIC_REGRESSION_CLASSIFIER, Modeller.GAM_CLASSIFIER, Modeller.RF_CLASSIFIER,
                    Modeller.GBM_CLASSIFIER, Modeller.LIGHTGBM_CLASSIFIER, Modeller.XGBOOST_CLASSIFIER,
                    Modeller.MULTI_LAYER_PERCEPTRON_CLASSIFIER]:
            return ModellerType.CLASSIFIER
        return ModellerType.REGRESSOR



"""Includes DqCategoricalStatistics class"""

from typing import List, Dict

from organon.idq.domain.businessobjects.statistics.dq_base_statistics import DqBaseStatistics
from organon.idq.domain.helpers.statistics.dq_statistical_functions import DqStatisticalFunctions


class DqCategoricalStatistics(DqBaseStatistics):
    """todo"""

    def __init__(self, missing_values: List[object], frequencies: Dict[object, float] = None):
        super().__init__()
        if frequencies is None:
            frequencies = {}
        self.missing_values: List[object] = missing_values
        self.frequencies: Dict[object, float] = frequencies
        self.cardinality = 0
        self.n_miss = 0
        self.n_val = 0

    @property
    def percentages(self) -> Dict[object, float]:
        """todo"""
        return DqStatisticalFunctions.get_percentages_from_frequencies(self.frequencies)

    def add(self, key: object, frequency: float):
        """Adds frequency value to statistics"""
        if key in self.frequencies:
            self.frequencies[key] += frequency
        else:
            self.frequencies[key] = frequency
            if key not in self.missing_values:
                self.cardinality += 1
        if key in self.missing_values:
            self.n_miss += frequency
        self.n_val += frequency



"""Includes SourceWithDateDqOutputReport class."""
import pandas as pd
from organon.idq.domain.reporting.objects.base_dq_output_report import BaseDqOutputReport


class SourceWithPartitionsDqOutputReport(BaseDqOutputReport):
    """Output report for DQ execution."""

    def __init__(self):
        super().__init__()
        self.table_control_details_df: pd.DataFrame = None
        self.numeric_column_control_details_df: pd.DataFrame = None
        self.nominal_column_control_details_df: pd.DataFrame = None



"""Includes SweepSelection class."""
from typing import Optional

import pandas as pd

from organon.ml.feature_selection.domain.objects.settings.sweep_selection_settings import SweepSelectionSettings
from organon.ml.feature_selection.domain.objects.sweep_selection_output import SweepSelectionOutput
from organon.ml.feature_selection.domain.services.sweep_selection_service import SweepSelectionService
from organon.ml.feature_selection.services.abstractions.base_unsupervised_feature_selecter import \
    BaseUnsupervisedFeatureSelecter


class SweepSelection(BaseUnsupervisedFeatureSelecter):
    """Feature selecter which utilizes sweep algorithm for selection"""

    def __init__(self, data: pd.DataFrame, *, bin_count: int = None, r_factor: float = None, random_state: int = None,
                 max_col_count: int = None, n_threads: int = 1):
        super().__init__(data)
        self.settings = SweepSelectionSettings(data, bin_count=bin_count, r_factor=r_factor, random_state=random_state,
                                               max_col_count=max_col_count, n_threads=n_threads)
        self.output: Optional[SweepSelectionOutput] = None

    def _run(self):
        service = SweepSelectionService()
        self.output = service.run_selection(self.settings)
        return self.output



"""Includes RFClassifier class."""
from sklearn.ensemble import RandomForestClassifier

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_classifier import BaseSklearnClassifier


class RFClassifier(BaseSklearnClassifier[RandomForestClassifier]):
    """Wrapper for sklearn RandomForestClassifier"""

    @property
    def classifier_class(self):
        return RandomForestClassifier



"""Includes UnlabeledDirDataSourceHandler class"""
import os

from pandas import DataFrame

from organon.ml.object_classification.common.object_classification_constants import \
    ObjectClassificationConstants
from organon.ml.object_classification.domain.data_sources.base_data_source_handler import \
    BaseDataSourceHandler


class UnlabeledDirDataSourceHandler(BaseDataSourceHandler[str]):
    """Class for handling Unlabeled Directory input"""

    def process_data(self):
        import tensorflow as tf # pylint: disable=import-outside-toplevel

        def parse_image(filename):
            parts = tf.strings.split(filename, os.sep)
            label = parts[-1]
            image = tf.io.read_file(filename)
            image = tf.io.decode_jpeg(image)
            return image, label

        list_ds = tf.data.Dataset.list_files(os.path.join(self.data, "*.*"))
        self._dataset = list_ds.map(parse_image)
        self._transform_tf_dataset()

    def initialize_probs_df(self) -> DataFrame:
        return DataFrame({ObjectClassificationConstants.FILENAME_COL_NAME: self.get_labels()})

    def _get_binary_labels(self):
        return [label.numpy().decode() for label in self._extract_labels()]

    def _get_multiclass_labels(self):
        return [label.numpy().decode() for label in self._extract_labels()]



"""
This module includes AfeStaticObjects class.
"""
from datetime import datetime

from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.fl.security.helpers.encryption_helper import EncryptionHelper


class AfeStaticObjects:
    """
    This class has static attributes corresponding to AFE constants.
    """
    distinct_entities_table_name_prefix: str = None
    distinct_entities_entity_column_name: str = None
    event_date_column_name: str = None
    empty_date_col_prefix: str = None
    no_date_col_default_date: datetime = None

    empty_quantity_column: str = None
    empty_dimension_column: str = None
    empty_dimension_val: str = None
    empty_dimension_index: int = None

    log_for_debugging: bool = None
    enable_trace_logging: bool = None
    __SECRET_KEY_DEFAULT = "qwertyuıop"  # nosec
    __secret_key: str = __SECRET_KEY_DEFAULT

    @staticmethod
    def set_defaults():
        """Sets default values."""
        AfeStaticObjects.distinct_entities_table_name_prefix = "DET"
        AfeStaticObjects.distinct_entities_entity_column_name = "ENTITY_ID"
        AfeStaticObjects.event_date_column_name = "EVENT_DATE"
        AfeStaticObjects.empty_date_col_prefix = "__DATE"
        AfeStaticObjects.no_date_col_default_date = datetime(1971, 1, 1)

        AfeStaticObjects.empty_quantity_column = "NoQuantity"
        AfeStaticObjects.empty_dimension_column = "NoDimension"
        AfeStaticObjects.empty_dimension_val = "NULL"
        AfeStaticObjects.empty_dimension_index = 0

        AfeStaticObjects.log_for_debugging = False

    @staticmethod
    def get_secret_key():
        """Returns value of secret key"""
        secret_key = AfeStaticObjects.__secret_key
        if secret_key == AfeStaticObjects.__SECRET_KEY_DEFAULT:
            LogHelper.warning("Secret key not set. Will use default secret key. Set secret key for security.")
        return secret_key

    @staticmethod
    def set_secret_key(secret_key: str):
        """Sets global encryption secret key for AFE"""
        if len(secret_key) > EncryptionHelper.SECRET_KEY_MAX_LENGTH:
            raise KnownException(f"secret_key maximum length is {EncryptionHelper.SECRET_KEY_MAX_LENGTH}")
        AfeStaticObjects.__secret_key = secret_key



"""Module for SweepOperator"""
from typing import Tuple, Iterable

import numpy as np
# pylint: disable=no-name-in-module
from organon.fl.mathematics.sweep.pyx_files.sweep_helper import sweep_fast_row_c, update_row_k

from organon.fl.mathematics.linearalgebra.IBlas import IBlas
from organon.fl.mathematics.sweep.data_classes.swept_attribute import SweptAttribute


class SweepOperator:
    """This is the main part of the algorithm. See sweep operator for further details.
    ftp://public.dhe.ibm.com/software/analytics/spss/support/Stats/Docs/Statistics/Algorithms/13.0/regression.pdf"""

    def __init__(self, matrix: np.ndarray, sigma: np.ndarray, target: int, blas_service: IBlas):
        if matrix.shape[0] != matrix.shape[1]:
            raise ValueError("Input matrix must be square")

        self._matrix = matrix
        self._sigma = sigma
        self._target = target

        self._length = matrix.shape[1]

        self._swept_indices = []
        self._nonswept_indices = []
        self._swept_ever = set()
        self._swept_never = set()
        self._swept_list = []
        self.reset()

        self._tmp_row = np.empty(shape=(self._length,), dtype=np.single)
        self._tmp_col = np.empty(shape=(self._length,), dtype=np.single)
        self.blas_service = blas_service

    def reset(self):
        """Resets sweep related lists."""
        self._swept_indices.clear()
        self._nonswept_indices.clear()
        self._nonswept_indices.extend(range(self._length))
        self._swept_never.update(range(self._length))
        self._nonswept_indices.remove(self._target)
        self._swept_never.remove(self._target)
        self._swept_list.clear()

    def add(self, k: int):
        """Adds element to sweep lists"""
        self._swept_indices.append(k)
        self._nonswept_indices.remove(k)
        self._swept_ever.add(k)
        if k in self._swept_never:
            self._swept_never.remove(k)

    def remove(self, k: int):
        """Removes elements from sweep lists"""
        self._swept_indices.remove(k)
        self._nonswept_indices.append(k)

    def sweep(self, k: int):
        """This is the modified sweep algorithm by Ömer Yılmaz. Still needs improvements. It will be done later.
        Basically it sweeps every attribute is ever swept if the attribute is not in the current swept list.
        Advantage: Reduces the cost of the including the attribute again and sweep the whole matrix again.
        Disadvantage: If an attribute won't be swept ever, this algorithm calculates it also every time."""
        is_in_swept, is_ever_swept = self.find_index_of(k)

        if is_ever_swept:
            # print("row, j = {}, {}".format(row, j))
            # print("Burası")
            attribute = self.get_swept_attribute_before_sweep(k)
            self.sweep_fast_row(self._matrix, k, self._target, self._swept_ever, self._length)
            self.set_swept_attribute_after_sweep(attribute, k)

            if is_in_swept:
                self.remove(k)
            else:
                self.add(k)
            self._swept_list.append(attribute)

        else:
            self.blas_service.scopy(self._length, self._matrix[k, :], 1, self._tmp_row, 1)
            # Operating on columns are a bit tricky via MKL Functions for C_contiguous arrays
            self.blas_service.scopy(self._length, self._matrix.ravel()[k:], self._length, self._tmp_col, 1)
            # self._tmp_col = self._matrix[:, row].copy()

            for attribute in self._swept_list:
                row = attribute.swept_row
                a_row = attribute.unswept_col[k]

                self.blas_service.saxpy(self._length, a_row, row, 1, self._tmp_row, 1)

                col = attribute.swept_col
                a_col = -1.0 * attribute.unswept_row[k]

                self.blas_service.saxpy(self._length, a_col, col, 1, self._tmp_col, 1)

            self.add(k)

            # print(self._tmp_row)
            # print(self._tmp_col)
            for index in self._swept_never:
                self._matrix[k, index] = self._tmp_row[index]
                self._matrix[index, k] = self._tmp_col[index]
            attribute = self.get_swept_attribute_before_sweep(k)
            self.sweep_fast_row(self._matrix, k, self._target, self._swept_ever, self._length)
            self.set_swept_attribute_after_sweep(attribute, k)
            self._swept_list.append(attribute)

    def sweep_old(self, k: int):
        """Old sweep algorithm from .net platform. Whenever an attribute excluded, it resets the matrix and
        calculates again"""
        # for i in range(len(self._swept_list)):
        #    print("i, index = {}, {}".format(i, self._swept_list[i].attribute_index))

        if len(self._swept_list) == 0:
            self.add(k)
            swept_attribute = self.get_swept_attribute_before_sweep(k)
            self.sweep_fast_row(self._matrix, k, self._target, self._swept_indices, self._length)
            self.set_swept_attribute_after_sweep(swept_attribute, k)
            self._swept_list.append(swept_attribute)

        else:
            is_in_swept, _ = self.find_index_of(k)

            if is_in_swept:
                # print("row, j = {}, {}".format(row, j))
                # print("Burası")
                for index in self._swept_indices:
                    self.sweep_fast_row(self._matrix, index, self._target, self._swept_indices, self._length)
                self._swept_indices.remove(k)
                tmp_list = self._swept_indices.copy()
                self.reset()
                for i in tmp_list:
                    self.sweep_old(i)

            else:
                self.blas_service.scopy(self._length, self._matrix[k, :], 1, self._tmp_row, 1)
                # Operating on columns are a bit tricky via MKL Functions for C_contiguous arrays
                self.blas_service.scopy(self._length, self._matrix.ravel()[k:], self._length, self._tmp_col, 1)
                # self._tmp_col = self._matrix[:, row].copy()

                for attribute in self._swept_list:
                    row = attribute.swept_row
                    a_row = attribute.unswept_col[k]

                    self.blas_service.saxpy(self._length, a_row, row, 1, self._tmp_row, 1)

                    col = attribute.swept_col
                    a_col = -1.0 * attribute.unswept_row[k]

                    self.blas_service.saxpy(self._length, a_col, col, 1, self._tmp_col, 1)

                self.add(k)

                # print(self._tmp_row)
                # print(self._tmp_col)

                for index in self._nonswept_indices:
                    self._matrix[k, index] = self._tmp_row[index]
                    self._matrix[index, k] = self._tmp_col[index]
                attribute = self.get_swept_attribute_before_sweep(k)
                self.sweep_fast_row(self._matrix, k, self._target, self._swept_indices, self._length)
                self.set_swept_attribute_after_sweep(attribute, k)
                self._swept_list.append(attribute)

    def get_swept_attribute_before_sweep(self, k: int):
        """Initialize swept attribute with non sweep row and col"""
        row = self._matrix[k, :].copy()
        col = self._matrix[:, k].copy()
        return SweptAttribute(attribute_index=k, unswept_row=row, unswept_col=col)

    def set_swept_attribute_after_sweep(self, attribute: SweptAttribute, k: int):
        """Sets the attribute's swept row and col"""
        attribute.swept_row = self._matrix[k, :].copy()
        attribute.swept_col = self._matrix[:, k].copy()

    def find_index_of(self, k: int) -> Tuple[bool, bool]:
        """Returns if attribute is currently swept, ever swept"""
        for val in self._swept_indices:
            if val == k:
                return True, True

        if k in self._swept_ever:
            return False, True
        return False, False


    def sweep_fast_row(self, matrix: np.ndarray, row: int, target: int, swept: Iterable[int], length: int):
        """Does the sweep algorithm along a row"""
        diag_value = matrix[row, row]
        self.blas_service.sscal(length, -1.0 / diag_value, matrix[row, :], 1)

        for i in range(matrix.shape[0]):
            if i == row:
                continue
            i_value = matrix[i, row]

            if i == target or i in swept:
                self.blas_service.saxpy(length, i_value, matrix[row, :], 1, matrix[i, :], 1)

            else:
                for j in swept:
                    matrix[i, j] += i_value * matrix[row, j]

                matrix[i, target] += i_value * matrix[row, target]
                matrix[i, i] += i_value * matrix[row, i]

            matrix[i, row] = i_value / diag_value
        matrix[row, row] = 1.0 / diag_value

    # @staticmethod
    # def sweep_fast_col(m, row, target, swept, length):
    #     D = m[row, row]
    #     print("m[{0}, {0}] = {1}".format(row, m[row, row]))
    #     print("target =", target)
    #     m_flat = m.ravel()
    #     col_n = m.shape[1]
    #     m[:, row] /= D
    #     swept = list(swept)
    #
    #     for i in range(col_n):
    #         if i == row:
    #             continue
    #         B = -m[row, i]
    #
    #         if i == target or i in swept:
    #             self.blas_service.daxpy(length, B, m_flat[row:], col_n, m_flat[i:], col_n)
    #
    #         else:
    #             self.blas_service.daxpy(len(swept), B, m[swept, row], 1, m[swept, i], 1)
    #
    #         m[row, i] = B / D
    #
    #         if np.isnan(m[1, 1]):
    #             print("{} ise sıkıntı".format(i))
    #     m[row, row] = 1.0 / D


class UnsupervisedSweepOperator:
    """
    Simpler sweep operator for unsupervised feature selection
    """

    def __init__(self, matrix: np.ndarray, blas_service: IBlas):
        if matrix.shape[0] != matrix.shape[1]:
            raise ValueError("Input matrix must be square")

        self._matrix = matrix
        if matrix.dtype not in (np.single, np.double):
            raise ValueError("Input matrix must be single or double")
        self.blas_service = blas_service
        if matrix.dtype.type == np.single:
            self._scal = self.blas_service.sscal
            self._copy = self.blas_service.scopy
            self._axpy = self.blas_service.saxpy
        else:
            self._scal = self.blas_service.dscal
            self._copy = self.blas_service.dcopy
            self._axpy = self.blas_service.daxpy

        self._length = matrix.shape[1]

        self._swept_ever = np.zeros((self._length,), dtype=np.uint8)
        self._swept_never = set(range(self._length))
        self._swept_list = []

        self._tmp_row = np.empty(shape=(self._length,), dtype=matrix.dtype)
        self._tmp_col = np.empty(shape=(self._length,), dtype=matrix.dtype)

    def add(self, k: int):
        """Adds element to sweep lists"""
        self._swept_ever[k] = True
        self._swept_never.remove(k)

    def sweep(self, k: int, num_threads: int):
        """Simpler sweep part"""
        self._tmp_row = np.copy(self._matrix[k, :])
        # self._copy(self._length, self._matrix[k, :], 1, self._tmp_row, 1)
        # Operating on columns are a bit tricky via MKL Functions for C_contiguous arrays
        self._tmp_col = np.copy(self._matrix[:, k])
        # self._copy(self._length, self._matrix.ravel()[k:], self._length, self._tmp_col, 1)
        # self._tmp_col = self._matrix[:, row].copy()

        for attribute in self._swept_list:
            row = attribute.swept_row
            a_row = attribute.unswept_col[k]

            # self._axpy(self._length, a_row, row, 1, self._tmp_row, 1)
            self._tmp_row += a_row * row

            col = attribute.swept_col
            a_col = -1.0 * attribute.unswept_row[k]

            # self._axpy(self._length, a_col, col, 1, self._tmp_col, 1)
            self._tmp_col += a_col * col

        self.add(k)

        # print(self._tmp_row)
        # print(self._tmp_col)
        update_row_k(self._matrix, self._tmp_row, self._tmp_col, self._swept_ever, k, num_threads)
        # for index in self._swept_never:
        #     self._matrix[k, index] = self._tmp_row[index]
        #     self._matrix[index, k] = self._tmp_col[index]
        attribute = self.get_swept_attribute_before_sweep(k)
        # self.sweep_fast_row_targetless(k)
        sweep_fast_row_c(self._matrix, self._length, self._swept_ever, k, num_threads)

        self.set_swept_attribute_after_sweep(attribute, k)
        self._swept_list.append(attribute)

    def get_swept_attribute_before_sweep(self, k: int):
        """Initialize swept attribute with non sweep row and col"""
        # TODO: we can fasten copying process using MKL
        row = self._matrix[k, :].copy()
        col = self._matrix[:, k].copy()
        return SweptAttribute(attribute_index=k, unswept_row=row, unswept_col=col)

    def set_swept_attribute_after_sweep(self, attribute: SweptAttribute, k: int):
        """Sets the attribute's swept row and col"""
        attribute.swept_row = self._matrix[k, :].copy()
        attribute.swept_col = self._matrix[:, k].copy()

    def sweep_fast_row_targetless(self, row: int):
        """Does the sweep algorithm along a row"""
        matrix = self._matrix
        diag_value = matrix[row, row]
        self._scal(self._length, -1.0 / diag_value, matrix[row, :], 1)

        for i in range(matrix.shape[0]):
            if i == row:
                continue
            i_value = matrix[i, row]

            if self._swept_ever[i]:
                self._axpy(self._length, i_value, matrix[row, :], 1, matrix[i, :], 1)

            else:
                for j in range(self._length):
                    if self._swept_ever[j]:
                        matrix[i, j] += i_value * matrix[row, j]
                matrix[i, i] += i_value * matrix[row, i]

            matrix[i, row] = i_value / diag_value
        matrix[row, row] = 1.0 / diag_value

    @property
    def matrix(self):
        """
        Returns the matrix
        :return: matrix
        """
        return self._matrix



"""Includes SelectionService class."""
import threading
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd

from organon.fl.core.executionutil import parallel_execution_helper
from organon.fl.core.executionutil.objects.locked_iterator import LockedIterator
from organon.fl.mathematics.constants import DOUBLE_MIN
from organon.ml.common.enums.target_type import TargetType
from organon.ml.common.helpers import validation_helper
from organon.ml.modelling.algorithms.classifiers.gam_classifier import GamClassifier
from organon.ml.modelling.algorithms.classifiers.rf_classifier import RFClassifier
from organon.ml.modelling.algorithms.classifiers.stacking_ensemble_classifier import StackingEnsembleClassifier
from organon.ml.modelling.algorithms.classifiers.voting_ensemble_classifier import VotingEnsembleClassifier
from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller
from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType
from organon.ml.modelling.algorithms.regressors.gam_regressor import GamRegressor
from organon.ml.modelling.algorithms.regressors.stacking_ensemble_regressor import StackingEnsembleRegressor
from organon.ml.modelling.algorithms.regressors.voting_ensemble_regressor import VotingEnsembleRegressor
from organon.ml.modelling.model_selection.domain.objects.selecter_output import SelecterOutput
from organon.ml.modelling.model_selection.domain.services.cross_validation_service import CrossValidationService
from organon.ml.modelling.model_selection.settings.objects.selection_settings import SelectionSettings


class SelectionService:
    """Service for selecting best modeller from a list of modellers"""

    @classmethod
    def select(cls, settings: SelectionSettings) -> SelecterOutput:
        """Selects best modeller from a list of modellers"""
        modeller_type = cls._run_initial_validation(settings)
        use_cv = settings.test_data is None
        fitted_stacking_modeller = None
        fitted_voting_modeller = None
        fitted_modellers = None
        if use_cv:
            pds = CrossValidationService.get_predefined_split(settings.train_data, settings.target_data,
                                                              modeller_type, cv_count=settings.cv_count)
            split_generator = LockedIterator(pds.split())
            all_scores = [None] * settings.cv_count
            if settings.num_threads_for_parallel_cv is not None:
                num_threads = cls.__get_num_threads(settings.num_threads_for_parallel_cv, settings.cv_count)
                params = [(all_scores, i, settings, split_generator) for i in range(settings.cv_count)]
                parallel_execution_helper.execute_parallel(params, cls._get_scores_for_split, num_jobs=num_threads,
                                                           require_shared_memory=True)
            else:
                for i in range(settings.cv_count):
                    cls._get_scores_for_split(all_scores, i, settings, split_generator)
        else:
            scores, fitted_modellers, fitted_stacking_modeller, fitted_voting_modeller = \
                cls._get_scores(settings, settings.train_data,
                                settings.target_data,
                                settings.test_data,
                                settings.test_target_data)
            all_scores = [scores]

        output = SelecterOutput()
        output.modeller_fold_scores = np.array([scores["modellers"] for scores in all_scores]).T.tolist()
        output.modeller_scores = list(np.mean(output.modeller_fold_scores, axis=1))

        stacking_mean_score, voting_mean_score = cls._fill_output_for_stacking_and_voting(all_scores, output,
                                                                                          settings.add_stacking,
                                                                                          settings.add_voting)
        all_mean_scores = output.modeller_scores + [stacking_mean_score, voting_mean_score]

        output.best_modeller = cls._get_best_modeller_as_fitted(settings, all_mean_scores, fitted_modellers,
                                                                fitted_stacking_modeller,
                                                                fitted_voting_modeller)

        return output

    @classmethod
    def _get_scores_for_split(cls, all_scores, i, settings: SelectionSettings, split_generator):
        train_indices, test_indices = next(split_generator)
        scores, _, _, _ = cls._get_scores(settings,
                                          settings.train_data.iloc[train_indices],
                                          settings.target_data.iloc[train_indices],
                                          settings.train_data.iloc[test_indices],
                                          settings.target_data.iloc[test_indices])
        all_scores[i] = scores

    @classmethod
    def _run_initial_validation(cls, settings: SelectionSettings):
        if settings.cv_count is None and settings.test_data is None:
            raise ValueError("Give a cv_count or test_data")

        modeller_type = settings.modellers[0].modeller_type
        if [modeller for modeller in settings.modellers if modeller.modeller_type != modeller_type]:
            raise ValueError("All modellers should be of same type")
        if modeller_type not in [ModellerType.CLASSIFIER, ModellerType.REGRESSOR]:
            raise ValueError("Only classifiers and regressors are supported in selection service")

        validation_helper.validate_get_target_type(modeller_type, settings.target_type)
        return modeller_type

    @classmethod
    def _fill_output_for_stacking_and_voting(cls, all_scores, output: SelecterOutput, add_stacking: bool,
                                             add_voting: bool):
        stacking_mean_score = DOUBLE_MIN
        if add_stacking:
            output.stacking_fold_scores = [scores["stacking"] for scores in all_scores]
            stacking_mean_score = np.mean(np.array(output.stacking_fold_scores))
            output.stacking_score = stacking_mean_score

        voting_mean_score = DOUBLE_MIN
        if add_voting:
            output.voting_fold_scores = [scores["voting"] for scores in all_scores]
            voting_mean_score = np.mean(np.array(output.voting_fold_scores))
            output.voting_score = voting_mean_score
        return stacking_mean_score, voting_mean_score

    @classmethod
    def _fit_and_append_model(cls, fitted_modellers, modeller, i, train_data, target_data):
        modeller = modeller.clone()
        modeller.fit(train_data, target_data)
        fitted_modellers[i] = modeller

    @classmethod
    def _get_best_modeller_as_fitted(cls, settings: SelectionSettings, all_mean_scores: List[float],
                                     fitted_modellers: Optional[List[BaseModeller]],
                                     fitted_stacking_modeller: Optional[BaseModeller],
                                     fitted_voting_modeller: Optional[BaseModeller]) -> BaseModeller:
        modeller_type = settings.modellers[0].modeller_type
        max_score = max(all_mean_scores)
        max_score_index = all_mean_scores.index(max_score)
        is_stacking_best = max_score_index == len(all_mean_scores) - 2
        is_voting_best = max_score_index == len(all_mean_scores) - 1

        cv_used = fitted_modellers is None

        if not cv_used:
            if is_stacking_best:
                return fitted_stacking_modeller
            if is_voting_best:
                return fitted_voting_modeller
            return fitted_modellers[max_score_index]

        best_modeller = None

        if is_stacking_best or is_voting_best:
            # if cv is used, modellers should be fitted with whole train data to be used as
            # estimators in stacking or voting
            fitted_modellers = [None] * len(settings.modellers)

            if settings.num_threads_for_parallel_fit is not None:
                num_threads = cls.__get_num_threads(settings.num_threads_for_parallel_fit, len(settings.modellers))
                params = [(fitted_modellers, modeller, i, settings.train_data, settings.target_data)
                          for i, modeller in enumerate(settings.modellers)]
                parallel_execution_helper.execute_parallel(params, cls._fit_and_append_model, num_jobs=num_threads,
                                                           require_shared_memory=True)
            else:
                for i, modeller in enumerate(settings.modellers):
                    cls._fit_and_append_model(fitted_modellers, modeller, i, settings.train_data, settings.target_data)

            if is_stacking_best:
                best_modeller = cls._get_stacking_modeller(fitted_modellers, modeller_type, settings.target_type)
                best_modeller.fit(settings.train_data, settings.target_data)

            elif is_voting_best:
                best_modeller = cls._get_voting_modeller(fitted_modellers, modeller_type)
                best_modeller.fit(settings.train_data, settings.target_data)
        else:
            best_modeller = settings.modellers[max_score_index].clone().fit(settings.train_data,
                                                                            settings.target_data)

        return best_modeller

    @classmethod
    def _fit_and_score_models(cls, fitted_modellers: list, modeller: BaseModeller,
                              fit_data: Tuple[pd.DataFrame, pd.Series],
                              score_data: Tuple[pd.DataFrame, pd.Series],
                              modeller_scores_for_fold: list, add_voting: bool,
                              all_predictions_df: pd.DataFrame, modeller_index: int, df_write_lock: threading.Lock):
        # pylint: disable=too-many-arguments
        modeller = modeller.clone()
        modeller.fit(fit_data[0], fit_data[1])
        fitted_modellers[modeller_index] = modeller

        score = modeller.score(score_data[0], score_data[1])
        modeller_scores_for_fold[modeller_index] = score
        if add_voting:
            prediction_df = modeller.predict(score_data[0])
            with df_write_lock:
                all_predictions_df[f"mdl_{modeller_index}"] = prediction_df[prediction_df.columns[0]].values

    @classmethod
    def _get_scores(cls, settings: SelectionSettings, train_data: pd.DataFrame, target_data: pd.Series,
                    test_data, test_target_data):
        all_predictions_df = pd.DataFrame()
        stacking_score = None
        voting_score = None
        modeller_type = settings.modellers[0].modeller_type
        fitted_modellers = [None] * len(settings.modellers)
        modeller_scores_for_fold = [None] * len(settings.modellers)
        df_write_lock = threading.Lock()
        if settings.num_threads_for_parallel_fit is not None:
            num_threads = cls.__get_num_threads(settings.num_threads_for_parallel_fit, len(settings.modellers))
            params = [(fitted_modellers, modeller, (train_data, target_data), (test_data, test_target_data),
                       modeller_scores_for_fold, settings.add_voting, all_predictions_df, i, df_write_lock)
                      for i, modeller in enumerate(settings.modellers)]
            parallel_execution_helper.execute_parallel(params, cls._fit_and_score_models, num_jobs=num_threads,
                                                       require_shared_memory=True)
        else:
            for i, modeller in enumerate(settings.modellers):
                cls._fit_and_score_models(fitted_modellers, modeller, (train_data, target_data),
                                          (test_data, test_target_data),
                                          modeller_scores_for_fold, settings.add_voting,
                                          all_predictions_df, i, df_write_lock)

        stacking_modeller = None
        if settings.add_stacking:
            stacking_modeller = cls._get_stacking_modeller(fitted_modellers,
                                                           modeller_type, settings.target_type)
            stacking_modeller.fit(train_data, target_data)
            stacking_score = stacking_modeller.score(test_data, test_target_data)

        voting_modeller = None
        if settings.add_voting:
            voting_modeller = cls._get_voting_modeller(fitted_modellers, modeller_type)
            voting_modeller.fit(train_data, target_data)
            voting_score = voting_modeller.score(test_data, test_target_data)

        return {"modellers": modeller_scores_for_fold, "stacking": stacking_score, "voting": voting_score}, \
               fitted_modellers, stacking_modeller, voting_modeller

    @classmethod
    def __get_num_threads(cls, num_threads: int, default=None):
        if num_threads == -1:
            if default is not None:
                return default
            return None
        return num_threads

    @classmethod
    def _get_stacking_modeller(cls, modellers, modeller_type: ModellerType, target_type: TargetType) -> BaseModeller:
        named_modellers = [(f"mdl_{i}", modeller) for i, modeller in enumerate(modellers)]
        if modeller_type == ModellerType.CLASSIFIER:
            final_estimator = RFClassifier() if target_type == TargetType.MULTICLASS else GamClassifier()
            return StackingEnsembleClassifier(estimators=named_modellers, final_estimator=final_estimator,
                                              cv="prefit", stack_method="predict_proba")
        return StackingEnsembleRegressor(estimators=named_modellers, final_estimator=GamRegressor(),
                                         cv="prefit")

    @classmethod
    def _get_voting_modeller(cls, modellers, modeller_type: ModellerType) -> BaseModeller:
        if modeller_type == ModellerType.CLASSIFIER:
            return VotingEnsembleClassifier(estimators=modellers, prefit=True)
        return VotingEnsembleRegressor(estimators=modellers, prefit=True)



"""Includes HPOptimizer class."""
from typing import List, Dict, Optional

import pandas as pd

from organon.ml.modelling.algorithms.services.ml_application_operations import MLApplicationOperations
from organon.ml.modelling.model_selection.domain.objects.hpo_output import HPOOutput
from organon.ml.modelling.model_selection.domain.services.cross_validation_service import DEFAULT_CV_COUNT
from organon.ml.modelling.model_selection.domain.services.hpo_service import HPOService
from organon.ml.modelling.model_selection.services.user_settings.user_hp_optimization_settings import \
    UserHPOptimizationSettings
from organon.ml.modelling.model_selection.settings.enums.search_type import SearchType
from organon.ml.modelling.model_selection.settings.model_selection_user_input_service import \
    ModelSelectionUserInputService


class HPOptimizer:
    """Hyper parameter optimizer"""

    def __init__(self):
        MLApplicationOperations.initialize_app()
        self.output = None

    def execute(self, train_data: pd.DataFrame, target_data: pd.Series, test_data: pd.DataFrame = None,
                test_target_data: pd.Series = None, cv_fold: int = DEFAULT_CV_COUNT,
                modellers: List[str] = None, modeller_params: List[Dict[str, list]] = None,
                search_method: str = SearchType.RANDOM.name, search_params: dict = None,
                scoring_metrics: Optional[List[str]] = None
                ) -> HPOOutput:
        """
        Execute optimizer and return output
        todo parametre açıklamaları
        """
        # pylint: disable=too-many-arguments

        user_settings = UserHPOptimizationSettings(train_data, target_data, search_method,
                                                   modellers=modellers, modeller_params=modeller_params,
                                                   cv_fold=cv_fold, test_data=test_data,
                                                   test_target_data=test_target_data,
                                                   scoring_metrics=scoring_metrics,
                                                   search_params=search_params)
        settings = ModelSelectionUserInputService().get_hp_optimization_settings(user_settings)
        self.output = HPOService().execute(settings)
        return self.output



"""
This module keeps the function used in ioc operations.
"""
from typing import Dict, TypeVar, Type

from organon.fl.core.exceptionhandling.known_exception import KnownException
from .ioc_registration_item import IocRegistrationItem
from ..businessobjects.fl_core_static_objects import FlCoreStaticObjects

IocAbstractType = TypeVar("IocAbstractType")


def register_type(item: IocRegistrationItem):
    """
    Checks if the item is already registered or not
    If so, raises a KnownException
    If not, registers the item
    :param item: IocRegistrationItem
    :return: nothing
    """
    if FlCoreStaticObjects.ioc_registered_items is None:
        FlCoreStaticObjects.ioc_registered_items = []

    for ioc_registered_item in FlCoreStaticObjects.ioc_registered_items:  # pylint: disable=not-an-iterable
        abstract_type_name = ioc_registered_item.abstract_type.__name__

        if ioc_registered_item.ioc_key == item.ioc_key and abstract_type_name == item.abstract_type.__name__:
            raise KnownException("Ioc key : " + ioc_registered_item.ioc_key + " abstract type : " + abstract_type_name +
                                 " is already registered")

    FlCoreStaticObjects.ioc_registered_items.append(item)


def resolve(abstract_type: Type[IocAbstractType], ioc_key: str, init_args: Dict = None, no_initialize=False, ) \
        -> IocAbstractType:
    """
    Resolves and returns the concrete type of the ioc registered item according to its abstract type and ioc key
    :param abstract_type: abstract type of the item
    :param ioc_key: ioc key of the item
    :param init_args: Creates an instance of concrete type with given arguments
    :param no_initialize: If true, does not create an instance and just returns concrete type
    :return: concrete type of the item
    """
    if FlCoreStaticObjects.ioc_registered_items is None:
        raise KnownException("No types registered")

    for ioc_registered_item in FlCoreStaticObjects.ioc_registered_items:  # pylint: disable=not-an-iterable
        abstract_type_name = ioc_registered_item.abstract_type.__name__

        if (ioc_registered_item.ioc_key == ioc_key) & (abstract_type_name == abstract_type.__name__):
            if not no_initialize:
                if init_args:
                    return ioc_registered_item.concrete_type(**init_args)
                return ioc_registered_item.concrete_type()
            return ioc_registered_item.concrete_type
    return None



"""
This module has functions validating attribute types of the classes required for AfeModeliingSettings
"""
import inspect
from datetime import datetime
from enum import Enum
from typing import List, Dict, get_args, get_origin

import numpy as np

from organon.afe.domain.enums.record_source_type import RecordSourceType
from organon.afe.domain.reporting.base_afe_model_output import BaseAfeModelOutput
from organon.afe.domain.settings.base_afe_modelling_settings import BaseAfeModellingSettings
from organon.afe.domain.settings.base_afe_scoring_settings import BaseAfeScoringSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.afe.domain.settings.validation_helper import cast_attr, raise_or_add_exception, cast_to_np_type
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers.date_helper import get_date_from_string

attr_basic_types = [str, int, float, bool, list, dict]
attr_np_types = [np.int64, np.short, np.float32, np.float64]


def attribute_types_validate(obj: object, attrs: Dict, obj_str: str = "", exception_list: bool = False) \
        -> List[Exception]:
    """
    Validates the types of attributes in the given object based on attribute types given in "attrs" dictionary.
    :param obj: Object instance to be validated
    :param attrs: A dictionary where keys are attribute names and values are expected types
    :param obj_str: A string that defines the given object
    :param exception_list: If true, this function will not raise an error and return a list of catched exceptions
    :return: Exception list.
    """
    exc_list = []
    for attr, expected_type in attrs.items():
        attr_val = getattr(obj, attr)
        val = __handle_type(attr_val, expected_type, f"{obj_str}.{attr}", exc_list, exception_list)
        setattr(obj, attr, val)
    return exc_list


def __handle_type(value, expected_type, definition: str, exc_list: List[Exception], add_to_list=False):
    if value is None:
        return None
    if inspect.isclass(expected_type):  # if expected_type is not a  "typing" type
        if expected_type in attr_basic_types or expected_type in attr_np_types:
            value = __handle_basic_or_np_type(value, expected_type, definition, exc_list, add_to_list=add_to_list)
        elif issubclass(expected_type, Enum):
            value = __handle_enum_type(value, expected_type, definition, exc_list, add_to_list=add_to_list)
        elif expected_type == datetime:
            value = __handle_datetime_type(value, definition, exc_list, add_to_list=add_to_list)
        else:
            exc_list.extend(_validate_types_for_class(value, expected_type, definition, exception_list=add_to_list))
    else:  # typing module types
        if get_origin(expected_type) == list:
            __handle_typing_list_type(value, expected_type, definition, exc_list, add_to_list=add_to_list)
        elif get_origin(expected_type) == dict:
            __handle_typing_dict_type(value, expected_type, definition, exc_list, add_to_list=add_to_list)
        else:
            raise NotImplementedError
    return value


def __handle_basic_or_np_type(value, expected_type, definition: str, exc_list: List[Exception], add_to_list=False):
    if expected_type in attr_basic_types or expected_type in attr_np_types:
        if not isinstance(value, expected_type):
            try:
                value = cast_to_basic_or_np(value, expected_type)
            except (ValueError, TypeError):
                exc = KnownException(
                    f"Type of '{definition}' is wrong. "
                    f"Expected: {expected_type} Got: {type(value)}(Value: {value})")
                raise_or_add_exception(exc, exc_list, add_to_list)
    return value


def __handle_datetime_type(value, definition: str, exc_list: List[Exception], add_to_list=False):
    if not isinstance(value, datetime):
        try:
            return get_date_from_string(value, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            exc = KnownException(
                f"Invalid date format or value in {definition}! Expected format:yyyy-mm-dd hh:mm:ss Got:{value}")
            raise_or_add_exception(exc, exc_list, add_to_list=add_to_list)
    return value


def __handle_typing_list_type(value, expected_type, definition: str, exc_list: List[Exception], add_to_list=False):
    list_element_type = get_args(expected_type)[0]
    if isinstance(value, list):
        for i in range(len(value)):  # pylint:disable=consider-using-enumerate
            elem = value[i]
            val = __handle_type(elem, list_element_type, f"{definition}[{i}]", exc_list, add_to_list=add_to_list)
            value[i] = val
    else:
        exc = KnownException(
            f"Type of '{definition}' is wrong. "
            f"Expected: List{list_element_type} Got: {type(value)}(Value: {value})")
        raise_or_add_exception(exc, exc_list, add_to_list)
    return value


def __handle_typing_dict_type(_dict, expected_type, definition: str, exc_list: List[Exception], add_to_list=False):
    key_type, value_type = get_args(expected_type)
    if isinstance(_dict, dict):
        for i, key in enumerate(list(_dict.keys())):
            value = _dict[key]
            new_key = __handle_type(key, key_type, f"{definition}.keys[{i}]", exc_list, add_to_list=add_to_list)
            new_value = __handle_type(value, value_type, f"{definition}[{key}]", exc_list, add_to_list=add_to_list)
            del _dict[key]
            _dict[new_key] = new_value
    else:
        exc = KnownException(
            f"Type of '{definition}' is wrong. "
            f"Expected: Dict[{key_type}, {value_type}] Got: {type(_dict)}(Value: {_dict})")
        raise_or_add_exception(exc, exc_list, add_to_list)
    return _dict


def cast_to_basic_or_np(value, value_type: type):
    """casts given value to given type if type in attr_basic_types or attr_np_types"""
    if value_type in attr_basic_types:
        return cast_attr(value, value_type)
    if value_type in attr_np_types:
        return cast_to_np_type(value, value_type)
    raise NotImplementedError


def __handle_enum_type(value, expected_enum_type, definition: str, exc_list: List[Exception], add_to_list=False):
    if isinstance(value, expected_enum_type):
        return value
    valid_values = [e.name for e in expected_enum_type]
    if value not in valid_values:
        exc = KnownException(
            f"Invalid value for enum '{expected_enum_type}' in {definition}. "
            f"Expected one of: {str(valid_values)} Got: {value}")
        raise_or_add_exception(exc, exc_list, add_to_list)
    else:
        value = expected_enum_type[value]
    return value


def _validate_types_for_class(obj, expected_type, obj_str: str, exception_list: bool = False) -> List[Exception]:
    """
    Validates the types of attributes in the given object

    :param obj: Object to be validated
    :param expected_type: Expected type for object
    :param obj_str: A string to define the object
    :type obj_str: str
    :param exception_list: If true, this function will not raise an error and return a list of catched exceptions
    :type exception_list: bool
    :return: Exception list
    """
    exc_list = []
    if not isinstance(obj, expected_type):
        exc = KnownException(
            f"Type of '{obj_str}' is wrong. "
            f"Expected: {expected_type} Got: {type(obj)}(Value: {obj})")
        raise_or_add_exception(exc, exc_list, exception_list)
    else:
        if issubclass(expected_type, BaseAfeModellingSettings):
            exc_list = validate_types_modelling_settings(obj, exception_list=exception_list)
        elif issubclass(expected_type, RecordSource):
            exc_list = validate_types_record_source(obj, obj_str, exception_list)
        elif hasattr(obj, "ATTR_DICT"):
            exc_list = validate_types_for_obj_with_attr_dict(obj, obj_str, exception_list)
        else:
            raise NotImplementedError(f"No type validation functions for given class: {expected_type}")

    return exc_list


def validate_types_for_obj_with_attr_dict(obj, obj_str: str = "", exception_list: bool = False) -> List[Exception]:
    """
    Validate types of attributes in obj instance using its ATTR_DICT attribute
    :param obj_str:
    :param obj:
    :param exception_list: If true, no exceptions will be raised and a list of catched exceptions will be returned
    :type exception_list: bool
    :return: Exception list
    """
    return attribute_types_validate(obj, obj.ATTR_DICT, obj_str, exception_list=exception_list)


def validate_types_modelling_settings(obj: BaseAfeModellingSettings, exception_list: bool = False) -> List[Exception]:
    """
    Validate types of attributes in AfeModellingSettings instance
    :param obj:
    :type obj: BaseAfeModellingSettings
    :param exception_list: If true, no exceptions will be raised and a list of catched exceptions will be returned
    :type exception_list: bool
    :return: Exception list
    """
    exc_list = validate_types_for_obj_with_attr_dict(obj, "", exception_list=exception_list)

    if obj.algorithm_settings is not None:
        exc_list.extend(validate_types_for_obj_with_attr_dict(obj.algorithm_settings, "algorithm_settings",
                                                              exception_list=exception_list))
    return exc_list


def validate_types_record_source(obj: RecordSource, obj_str: str = "record_source",
                                 exception_list: bool = False) -> List[Exception]:
    """
    Validate types of attributes in TargetRecordSource instance
    :param RecordSource obj:
    :param str obj_str: A string to define the instance of object
    :param bool exception_list: If true, no exceptions will be raised and a list of catched exceptions will be returned
    :return: Exception list
    """
    exc_list = []
    if obj.get_type() == RecordSourceType.DATABASE:
        exc_list.extend(validate_types_for_obj_with_attr_dict(obj.source, obj_str=f"{obj_str}.source",
                                                              exception_list=exception_list))
    return exc_list


def validate_types_scoring_settings(obj: BaseAfeScoringSettings, exception_list: bool = False) -> List[Exception]:
    """
    Validate types of attributes in AfeScoringSettings instance
    :param obj:
    :type obj: PAfeScoringSettings
    :param exception_list: If true, no exceptions will be raised and a list of catched exceptions will be returned
    :type exception_list: bool
    :return: Exception list
    """
    exc_list = []
    attrs = obj.ATTR_DICT.copy()
    attrs.pop("model_output")
    exc_list.extend(attribute_types_validate(obj, attrs, exception_list=exception_list))
    if obj.model_output is not None and not isinstance(obj.model_output, BaseAfeModelOutput):
        exc = KnownException("scoring_settings.model_output is not an instance of BaseAfeModelOutput")
        raise_or_add_exception(exc, exc_list, add_to_list=exception_list)

    return exc_list



"""Includes CoarseClassService class"""
from typing import Union, Tuple

import numpy as np
import pandas as pd

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.ml.common.enums.target_type import TargetType
from organon.ml.common.helpers.df_ops_helper import get_train_test_split_from_strata, get_sample_indices_from_series
from organon.ml.preprocessing.domain.objects.coarse_class_fit_output import CoarseClassFitOutput
from organon.ml.preprocessing.helpers.preprocessing_input_helper import convert_str_series_to_binary
from organon.ml.preprocessing.settings.objects.coarse_class_settings import CoarseClassSettings


class CoarseClassService:
    """Service for CoarseClass conversion"""

    def __init__(self, settings: CoarseClassSettings):
        if settings.target_type not in [TargetType.SCALAR, TargetType.BINARY]:
            raise ValueError("Target type should be scalar or binary.")
        self.settings = settings
        self.fit_output: CoarseClassFitOutput = None

    def fit(self, data: pd.DataFrame, target_data: pd.Series, stability_check_data: pd.DataFrame = None,
            stability_check_target_data: pd.Series = None) -> CoarseClassFitOutput:
        """fits coarse column and outputs coarse tables"""
        self._validate_data(data, target_data, stability_check_target_data, stability_check_target_data)
        target_native_type = get_column_native_type(pd.DataFrame({"target": target_data}), "target")
        if self.settings.target_type == TargetType.SCALAR and target_native_type != ColumnNativeType.Numeric:
            raise ValueError("Target column should be numeric for categorical coarse classing.")
        target_data, stability_check_target_data = self._check_target_data(target_data, stability_check_target_data,
                                                                           target_native_type)
        x_train, y_train, x_valid, y_valid = self._prepare_train_and_validation_data(data, target_data,
                                                                                     stability_check_data,
                                                                                     stability_check_target_data)
        fit_output = self._initialize_output(y_train)
        for col in x_train.columns:
            column_type = get_column_native_type(x_train, col)
            fit_output = self._execute_coarsing_for_col(fit_output, column_type, col, x_train[col], y_train,
                                                        x_valid[col], y_valid)
        self.fit_output = fit_output
        return fit_output

    def _validate_data(self, data: pd.DataFrame, target_data: pd.Series, stability_check_data: pd.DataFrame = None,
                       stability_check_target_data: pd.Series = None):
        if data is None or data.empty:
            raise ValueError("Data cannot be empty or None.")
        if self.settings.target_type == TargetType.BINARY and target_data.nunique() > 2:
            raise ValueError("Target type is binary but target contains more than two classes.")
        if len(data) != len(target_data) or any(data.index != target_data.index):
            raise ValueError("Data and target indexes are different.")
        if self.settings.stability_check:
            if stability_check_data is not None and stability_check_target_data is not None:
                if len(stability_check_data) != len(stability_check_target_data) or \
                        any(stability_check_data.index != stability_check_target_data.index):
                    raise ValueError("Stability check data and target data indexes are different.")

    def _check_target_data(self, target_data: pd.Series, stability_check_target_data: pd.Series,
                           target_native_type: ColumnNativeType):
        change_valid_data = False
        if self.settings.target_type == TargetType.BINARY and target_native_type != ColumnNativeType.Numeric:
            if self.settings.stability_check and (stability_check_target_data is not None):
                valid_target_type = get_column_native_type(pd.DataFrame(data=stability_check_target_data,
                                                                        columns=["target"]), "target")
                if valid_target_type == ColumnNativeType.Numeric:
                    raise ValueError("Target data includes non-numeric values but "
                                     "stability check data includes numeric values")
                change_valid_data = True
            if self.settings.positive_class is None or self.settings.negative_class is None:
                raise ValueError("Target includes non-numeric values but positive and negative classes are not given.")
            if change_valid_data:
                stability_check_target_data = convert_str_series_to_binary(
                    stability_check_target_data, self.settings.positive_class, self.settings.negative_class)
            target_data = convert_str_series_to_binary(target_data, self.settings.positive_class,
                                                       self.settings.negative_class)
        return target_data, stability_check_target_data

    def _execute_coarsing_for_col(self, fit_output: CoarseClassFitOutput, column_type: ColumnNativeType, col: str,
                                  x_train: pd.Series, y_train: pd.Series, x_valid: pd.Series, y_valid: pd.Series):
        # pylint: disable=too-many-arguments
        if x_train.nunique() <= 1:
            LogHelper.info(f"{col} cannot be coarse classed. Because it consists of a single value.")
            fit_output.rejected_list.append(col)
            return fit_output
        if column_type in [ColumnNativeType.Numeric, ColumnNativeType.String]:
            dt_df, x_val_tmp, y_val_tmp, null_char_table, null_cc_table, char_df = \
                self._summarize_column(x_train, y_train, x_valid, y_valid, col, column_type)
            if self._check_has_at_least_two_classes(dt_df):
                dt_df = self._coarse_column(dt_df, char_df, column_type, x_val_tmp, y_val_tmp,
                                            fit_output.target_mean)
                # if no stability check is desired then we are done here
                # continue for stability check
                if self.settings.stability_check:
                    dt_df = CoarseClassService._execute_stability(dt_df, self.settings.stability_threshold,
                                                                  column_type,
                                                                  fit_output.target_mean)
                fit_output.coarse_class_table = pd.concat([fit_output.coarse_class_table, dt_df, null_cc_table]
                                                          , ignore_index=True)
                fit_output.coarse_class_table["class_target_mean"].fillna(fit_output.target_mean, inplace=True)
                fit_output.char_table = pd.concat([fit_output.char_table, char_df, null_char_table],
                                                  ignore_index=True)
            else:
                LogHelper.info(f"{col} cannot be coarse classed.")
                fit_output.rejected_list.append(col)
        else:
            LogHelper.info(f"{col} cannot be coarse classed because its type is not numeric or object.")
            fit_output.rejected_list.append(col)
        return fit_output

    @staticmethod
    def _initialize_output(y_train: pd.Series) -> CoarseClassFitOutput:
        fit_output = CoarseClassFitOutput()
        fit_output.coarse_class_table = pd.DataFrame()
        fit_output.rejected_list = []
        fit_output.char_table = pd.DataFrame()
        # Save target mean and count
        fit_output.target_mean = y_train.mean()
        return fit_output

    @staticmethod
    def _check_has_at_least_two_classes(data_frame: pd.DataFrame) -> bool:
        return data_frame[data_frame['class_count'] > 0].shape[0] > 1

    def _summarize_column(self, x_train: pd.Series, y_train: pd.Series, x_valid: pd.Series,
                          y_valid: pd.Series, col: str, column_type: ColumnNativeType):
        x_tmp, y_tmp, x_val_tmp, y_val_tmp, y_nulls, y_valid_nulls = self._get_column_dfs(x_train,
                                                                                          y_train,
                                                                                          x_valid,
                                                                                          y_valid)
        null_char_table, null_cc_table = self._fill_cc_table_for_nulls(y_nulls, y_valid_nulls, col,
                                                                       y_train.mean(), y_train.count())
        char_df = pd.DataFrame()
        # convert categorical columns to numeric by assigning their target mean
        if column_type == ColumnNativeType.String:
            char_df, x_tmp = self._convert_categorical_col_to_numeric(x_tmp, col, y_tmp)
        # Build decision trees
        dt_df = self._build_decision_tree_df(self.settings.target_type, self.settings.min_class_size,
                                             self.settings.max_leaf_nodes, x_tmp, y_tmp, col)
        del x_tmp
        # Prepare coarse class summary table
        dt_df = self._prepare_cc_summary(dt_df, y_tmp, col, y_train.count())
        return dt_df, x_val_tmp, y_val_tmp, null_char_table, null_cc_table, char_df

    def _prepare_train_and_validation_data(self, x_train: pd.DataFrame, y_train: pd.Series,
                                           stability_check_data: pd.DataFrame, stability_target_data: pd.Series) \
            -> Tuple[pd.DataFrame, pd.Series, pd.DataFrame, pd.Series]:
        """prepares train and validation data sets"""
        # Get validation data for stability check
        if self.settings.stability_check:
            if stability_check_data is None:
                x_train, y_train, x_valid, y_valid = get_train_test_split_from_strata(x_train,
                                                                                      y_train,
                                                                                      self.settings.test_ratio,
                                                                                      self.settings.random_state)
            else:
                y_valid = stability_target_data
                x_valid = stability_check_data
        else:
            valid_indices = get_sample_indices_from_series(y_train, self.settings.test_ratio,
                                                           self.settings.random_state)
            x_valid, y_valid = x_train.loc[valid_indices], y_train.loc[valid_indices]
        return x_train, y_train, x_valid, y_valid

    @staticmethod
    def _get_column_dfs(x_train: pd.Series, y_train: pd.Series, x_valid: pd.Series, y_valid: pd.Series) -> Tuple[
        pd.Series, pd.Series, pd.Series, pd.Series, pd.Series, pd.Series]:
        y_nulls = y_train.loc[y_train.index.isin(x_train[x_train.isnull()].index)]
        y_valid_nulls = y_valid.loc[y_valid.index.isin(x_valid[x_valid.isnull()].index)]
        return x_train.dropna(), y_train.loc[y_train.index.isin(x_train.dropna().index)], \
               x_valid.dropna(), y_valid.loc[
                   y_valid.index.isin(x_valid.dropna().index)], y_nulls, y_valid_nulls

    @staticmethod
    def _fill_cc_table_for_nulls(y_nulls: pd.Series, y_valid_nulls: pd.Series, col, target_mean: float,
                                 class_count: int):
        y_null_mean = y_nulls.mean() if len(y_nulls) > 0 else float("nan")
        null_char_table = pd.DataFrame(
            {"char_value": ["NULL"],
             "num_value": [y_null_mean],
             "variable": [col]})
        null_char_table["num_value"].fillna(target_mean, inplace=True)
        null_cc_table = pd.DataFrame(
            {"class_min": [y_nulls.min()], "class_max": [y_nulls.max()], "class_count": [y_nulls.count()],
             "class_target_mean": [y_null_mean], "class_target_count": [y_nulls.sum()],
             "class_no": ["nulls"], "class_percent": [y_nulls.count() / class_count],
             "class_target_deviation": [
                 (y_null_mean - target_mean) / target_mean],
             "variable": [col], "class": ["null"],
             "class_count (val)": [y_valid_nulls.count()], "class_target_mean (val)": [y_valid_nulls.mean()],
             "class_target_count (val)": [y_valid_nulls.sum()],
             "unstability": [np.nan]})
        null_cc_table["class_target_mean"].fillna(target_mean, inplace=True)
        null_cc_table["class_target_mean (val)"].fillna(target_mean, inplace=True)
        return null_char_table, null_cc_table

    @staticmethod
    def _convert_categorical_col_to_numeric(x_train: pd.Series, column_name: str, y_train: pd.Series
                                            ) -> Tuple[pd.DataFrame, pd.Series]:
        """converts categorical columns to numeric columns"""
        data = pd.DataFrame(columns=[column_name, 'target'])
        data[column_name] = x_train
        data["target"] = y_train
        char_df = data.groupby(column_name).agg({"target": np.mean})
        char_df.reset_index(inplace=True)
        char_df.columns = [column_name, column_name + '_num']
        data['indexcol'] = data.index
        data = pd.merge(data, char_df, how='left', on=column_name, copy=False)
        data = data.drop([column_name, "target"], axis=1).rename(
            columns={column_name + '_num': column_name})
        data.set_index('indexcol', inplace=True)
        char_df.columns = ['char_value', 'num_value']
        char_df["num_value"] = char_df["num_value"]
        char_df['variable'] = column_name
        return char_df, data[column_name]

    @staticmethod
    def _build_decision_tree_df(target_type: TargetType, min_class_size: int, max_leaf_nodes: int,
                                x_fit_all: pd.Series, y_fit: pd.Series, column_name: str) -> pd.DataFrame:
        dt_df = pd.DataFrame(columns=[column_name, 'dt'])
        dt_df[column_name] = x_fit_all
        if target_type == TargetType.BINARY:
            from sklearn.tree import DecisionTreeClassifier  # pylint: disable=import-outside-toplevel
            dt_classifier = DecisionTreeClassifier(min_samples_leaf=min_class_size,
                                                   max_leaf_nodes=max_leaf_nodes)
            dt_classifier.fit(x_fit_all.values.reshape(-1, 1), y_fit.values.reshape(-1, 1))
            dt_df["dt"] = dt_classifier.predict_proba(x_fit_all.values.reshape(-1, 1))[:, 1]
        elif target_type == TargetType.SCALAR:
            from sklearn.tree import DecisionTreeRegressor  # pylint: disable=import-outside-toplevel
            dt_regressor = DecisionTreeRegressor(min_samples_leaf=min_class_size,
                                                 max_leaf_nodes=max_leaf_nodes)
            dt_regressor.fit(x_fit_all.values.reshape(-1, 1), y_fit.values.reshape(-1, 1))
            dt_df["dt"] = dt_regressor.predict(x_fit_all.values.reshape(-1, 1))
        return dt_df

    def _prepare_cc_summary(self, dt_df: pd.DataFrame, y_train: pd.Series,
                            column_name: str, class_count: int) -> pd.DataFrame:
        dt_df = self._prepare_cc_parameters_df(dt_df, y_train, column_name)
        dt_df.sort_values(by='class_min', inplace=True)
        dt_df.reset_index(inplace=True)
        dt_df['class_no'] = np.where(dt_df['class_min'].isnull(), 'nulls', dt_df.index + 1)
        dt_df['class_percent'] = dt_df['class_count'] / class_count
        dt_df['class_target_deviation'] = (dt_df['class_target_mean'] - y_train.mean()) / y_train.mean()
        dt_df['variable'] = column_name
        del dt_df["dt"]
        return dt_df

    def _prepare_cc_parameters_df(self, dt_df: pd.DataFrame, y_train: pd.Series, column_name: str) -> pd.DataFrame:
        dt_df_base = dt_df.copy()
        dt_df_base["target"] = y_train
        dt_df = self._prepare_class_statistics(dt_df_base, column_name)
        # if class boundaries overlap, then class intervals are reconstructed.
        while self._check_class_overlap(dt_df):
            min_value = dt_df[dt_df['lag'] < dt_df['class_max']]['lag'].values[0]
            dt_value = dt_df[dt_df['lag'] == min_value].index.values[0]
            dt_df_base.loc[(dt_df_base[column_name] > min_value) & (dt_df_base['dt'] == dt_value),
                           "dt"] = dt_df_base['dt'] + 0.0000000001
            dt_df = self._prepare_class_statistics(dt_df_base, column_name)
        dt_df["class_max"] = dt_df["class_min"].shift(-1)
        dt_df.loc[dt_df.index[-1], "class_max"] = np.inf
        dt_df.drop('lag', axis=1, inplace=True)
        return dt_df

    @staticmethod
    def _check_class_overlap(dt_df: pd.DataFrame):
        return dt_df[dt_df['lag'] < dt_df['class_max']]['lag'].count() > 0

    @staticmethod
    def _prepare_class_statistics(dt_df_base: pd.DataFrame, column_name: str) -> pd.DataFrame:
        dt_df = dt_df_base.groupby('dt').agg({column_name: [min, max, 'count'], "target": [np.mean, sum]})
        dt_df.columns = ['class_min', 'class_max', 'class_count', 'class_target_mean', 'class_target_count']
        dt_df.sort_values(by='class_min', inplace=True)
        dt_df['lag'] = dt_df['class_min'].shift(-1)
        return dt_df

    def _coarse_column(self, dt_df: pd.DataFrame, char_df: pd.DataFrame,
                       column_type: ColumnNativeType,
                       x_valid: pd.Series, y_valid: pd.Series,
                       target_mean: Union[float, int]) -> pd.DataFrame:
        # pylint: disable=too-many-arguments
        dt_df = self._get_categorical_class_parameters(dt_df, column_type, char_df)
        # first calculate class count, class target count and class target mean for validation data
        # it is done separately for numeric and non-numeric columns
        if column_type == ColumnNativeType.Numeric:
            dt_df_val = CoarseClassService._calculate_validation_parameters_numeric(dt_df, x_valid, y_valid)
        else:
            dt_df_val = CoarseClassService._calculate_validation_parameters_object(dt_df, x_valid, y_valid)
        dt_df_val.loc[dt_df_val['class_target_mean (val)'].isnull(), 'class_target_mean (val)'] = target_mean
        dt_df = pd.merge(dt_df, dt_df_val, how='left', on='class_no', copy=False)
        return dt_df

    @staticmethod
    def _get_categorical_class_parameters(dt_df: pd.DataFrame, column_type: ColumnNativeType,
                                          char_df: pd.DataFrame) -> pd.DataFrame:
        dt_df['class'] = round(dt_df['class_min'], 3).astype(str) + '-' + round(dt_df['class_max'], 3).astype(
            str)
        # if column is categorical, then find which class it belongs to and append category to its "class" field.
        if column_type == ColumnNativeType.String:
            char_df['dummy'] = 1
            dt_df['dummy'] = 1
            df_tmp = pd.merge(char_df, dt_df[['class_min', 'class_max', 'class', 'dummy']], how='outer',
                              on='dummy', copy=False)
            char_df.drop('dummy', axis=1, inplace=True)
            dt_df.drop('dummy', axis=1, inplace=True)
            df_tmp = df_tmp[
                (df_tmp['num_value'] >= df_tmp['class_min']) & (df_tmp['num_value'] < df_tmp['class_max'])]
            dt_df = pd.merge(dt_df, pd.DataFrame(
                df_tmp.groupby(['class'])['char_value'].apply(','.join)).reset_index(), on='class', copy=False).drop(
                'class', axis=1).rename(columns={'char_value': 'class'})
        return dt_df

    @classmethod
    def _calculate_validation_parameters_numeric(cls, dt_df: pd.DataFrame, x_valid: pd.Series,
                                                 y_valid: pd.Series) -> pd.DataFrame:
        # pylint: disable=cell-var-from-loop
        dt_df_val = dt_df[['class_no', 'class_count', 'class_target_mean', 'class_target_count']]
        dt_df_val.columns = ['class_no', 'class_count (val)', 'class_target_mean (val)',
                             'class_target_count (val)']
        for class_no in dt_df["class_no"]:
            minn = dt_df[dt_df["class_no"] == class_no]["class_min"].values[0]
            maxx = dt_df[dt_df["class_no"] == class_no]["class_max"].values[0]
            target_values = y_valid[
                y_valid.index.isin(x_valid.where(lambda x: (x >= minn) & (x < maxx)).dropna().index)]
            class_count = target_values.count()
            class_mean = target_values.mean()
            class_sum = target_values.sum()
            dt_df_val.loc[dt_df_val['class_no'] == class_no, 'class_count (val)'] = class_count
            dt_df_val.loc[dt_df_val['class_no'] == class_no, 'class_target_mean (val)'] = class_mean
            dt_df_val.loc[dt_df_val['class_no'] == class_no, 'class_target_count (val)'] = class_sum
        return dt_df_val

    @classmethod
    def _calculate_validation_parameters_object(cls, dt_df: pd.DataFrame, x_valid: pd.Series,
                                                y_valid: pd.Series) -> pd.DataFrame:
        x_valid.fillna('NULL', inplace=True)
        dt_df_val = dt_df[['class_no', 'class_count', 'class_target_mean', 'class_target_count']]
        dt_df_val.columns = ['class_no', 'class_count (val)', 'class_target_mean (val)',
                             'class_target_count (val)']
        for class_no in dt_df["class_no"].values:
            value_list = dt_df[dt_df['class_no'] == class_no]['class'].values[0].split(",")
            dt_df_val.loc[dt_df_val['class_no'] == class_no, 'class_count (val)'] = \
                x_valid[x_valid.isin(value_list)].count()
            dt_df_val.loc[dt_df_val['class_no'] == class_no, 'class_target_mean (val)'] = y_valid[
                y_valid.index.isin(x_valid[x_valid.isin(value_list)].index)].mean()
            dt_df_val.loc[dt_df_val['class_no'] == class_no, 'class_target_count (val)'] = y_valid[
                y_valid.index.isin(x_valid[x_valid.isin(value_list)].index)].sum()
        return dt_df_val

    @staticmethod
    def _assign_unstability(dt_df: pd.DataFrame) -> pd.DataFrame:
        dt_df['unstability'] = np.where(dt_df['class_target_mean'] != 0.0,
                                        round(abs(dt_df['class_target_mean'] - dt_df['class_target_mean (val)']) / (
                                                dt_df['class_target_mean'] + 0.0000000001), 2), 1.0)
        dt_df['unstability'] = np.where((dt_df['class_target_mean'] == 0.0) & (dt_df["class_target_mean (val)"] == 0.0),
                                        0, dt_df['unstability'])
        return dt_df

    @classmethod
    def _execute_stability(cls, dt_df: pd.DataFrame, stability_threshold: float,
                           column_type: ColumnNativeType,
                           target_mean: float) -> pd.DataFrame:
        dt_df = cls._assign_unstability(dt_df)
        # Loop until there is no unstable class
        while dt_df[dt_df['unstability'] > stability_threshold]['class_no'].count() > 0 and len(dt_df["class_no"]) > 1:
            max_class_no = dt_df['class_no'].astype(int).max()
            unstable_class, combine_class = cls._get_unstable_and_combine_class(dt_df, stability_threshold, column_type,
                                                                                max_class_no)
            # combine two classes
            combined_row = cls._combine_two_classes(dt_df, unstable_class, combine_class, target_mean)
            dt_df = dt_df[~dt_df['class_no'].isin([str(unstable_class), str(combine_class)])]
            dt_df = pd.concat([dt_df, combined_row])
            dt_df.sort_values(by='class_min', inplace=True)
            dt_df.reset_index(inplace=True, drop=True)
            dt_df['class_no'] = np.where(dt_df['class_no'] == 'nulls', 'nulls', dt_df.index + 1)
        return dt_df

    @staticmethod
    def _get_unstable_and_combine_class(dt_df: pd.DataFrame, stability_threshold: float, column_type: ColumnNativeType,
                                        max_class_no: int) -> Tuple[int, int]:
        unstable_class = dt_df[(dt_df['unstability'] > stability_threshold)][
            'class_no'].astype(int).min()
        # for numeric columns unstable classes are combined with their adjacent classes
        if column_type == ColumnNativeType.Numeric:
            if unstable_class == 1:
                combine_class = unstable_class + 1
            elif unstable_class == max_class_no:
                combine_class = unstable_class - 1
            else:
                t_mean = dt_df[dt_df['class_no'] == str(unstable_class)]['class_target_mean'].values[0]
                next_mean = \
                    dt_df[dt_df['class_no'] == str(unstable_class + 1)]['class_target_mean'].values[0]
                prev_mean = \
                    dt_df[dt_df['class_no'] == str(unstable_class - 1)]['class_target_mean'].values[0]
                if abs(t_mean - next_mean) <= abs(t_mean - prev_mean):
                    combine_class = unstable_class + 1
                else:
                    combine_class = unstable_class - 1
        # for non-numeric columns, an unstable class is combined with the one having the closes target mean
        else:
            unstable_class_mean = \
                dt_df[dt_df['class_no'] == str(unstable_class)]['class_target_mean'].values[0]
            min_difference = abs(dt_df[dt_df['class_no'] != str(unstable_class)][
                                     'class_target_mean'] - unstable_class_mean).min()
            dt_df['difference'] = abs(dt_df['class_target_mean'] - unstable_class_mean)
            combine_class = int(
                dt_df[dt_df['difference'] == min_difference]['class_no'].values[0])
            dt_df.drop('difference', axis=1, inplace=True)
        return unstable_class, combine_class

    @staticmethod
    def _combine_two_classes(dt_df: pd.DataFrame, unstable_class: int, combine_class: int,
                             target_mean: float) -> pd.DataFrame:
        unstable_row = dt_df[dt_df['class_no'] == str(unstable_class)]
        combine_row = dt_df[dt_df['class_no'] == str(combine_class)]
        combined_row = pd.concat([unstable_row, combine_row])
        new_row = combined_row.groupby('variable').agg({'class_min': min,
                                                        'class_max': max,
                                                        'class_count': sum,
                                                        'class_target_count': sum,
                                                        'class_no': min,
                                                        'class_percent': sum,
                                                        'class_count (val)': sum,
                                                        'class_target_count (val)': sum})
        new_row = pd.concat(
            [new_row, pd.DataFrame(combined_row.groupby(['variable'])['class'].apply(','.join))],
            axis=1)
        new_row['class_no'] = new_row['class_no'].astype(str)
        new_row['class_target_mean'] = new_row['class_target_count'] / new_row['class_count']
        new_row['class_target_mean (val)'] = new_row['class_target_count (val)'] / new_row[
            'class_count (val)']
        new_row['class_target_mean (val)'].fillna(target_mean, inplace=True)
        new_row.reset_index(inplace=True)
        new_row['class_target_deviation'] = (new_row['class_target_mean'] - target_mean) / target_mean
        CoarseClassService._assign_unstability(new_row)
        new_row = new_row[unstable_row.columns]
        return new_row

    def transform(self, data: pd.DataFrame, with_class_value=False):
        """transforms data"""
        if self.fit_output is None:
            raise ValueError("You should call fit method first to transform data.")
        base_df = data.drop(data.columns, axis=1)
        for col in data.columns:
            transformed_col_name = "Z_" + str(col)
            class_value_col_name = "C_" + str(col)
            column_type = get_column_native_type(data, col)
            if col not in self.fit_output.rejected_list:
                x_df = data[col]
                c_series_base = data[col]
                c_df = pd.DataFrame()
                c_df["C_col"] = x_df
                cc_table = self.fit_output.coarse_class_table[self.fit_output.coarse_class_table['variable'] == col]
                if column_type == ColumnNativeType.Numeric:
                    x_df, c_df = self._fill_outliers(x_df, c_series_base, c_df, cc_table, with_class_value)
                    x_df, c_df = self._fill_nulls(x_df, c_series_base, c_df, cc_table, with_class_value)
                    x_df, c_df = self._fill_for_not_objects(x_df, c_series_base, c_df, cc_table, with_class_value)
                else:
                    x_df, c_df = self._fill_for_objects(x_df, c_series_base, c_df, cc_table, with_class_value)
                base_df[transformed_col_name] = x_df
                if with_class_value:
                    base_df[class_value_col_name] = c_df["C_col"]
                del x_df
        return base_df

    def _fill_nulls(self, x_df: pd.Series, c_series: pd.Series, c_df: pd.DataFrame, cc_table: pd.DataFrame,
                    with_class_value: bool) -> Tuple[
        pd.Series, pd.DataFrame]:
        if with_class_value:
            c_df.loc[c_series.isnull(), "C_col"] = "null"
        return pd.Series(data=np.where(x_df.isnull(),
                                       cc_table[cc_table['class'] == "null"][
                                           'class_target_mean'] - self.fit_output.target_mean,
                                       x_df)), c_df

    @staticmethod
    def _fill_outliers(x_df: pd.Series, c_series: pd.Series, c_df: pd.DataFrame, cc_table: pd.DataFrame,
                       with_class_value: bool) -> Tuple[
        pd.Series, pd.DataFrame]:
        minn = cc_table[cc_table["class"] != "null"]['class_min'].min()
        maxx = np.inf if np.isinf(cc_table[cc_table["class"] != "null"]['class_max'].max()) else \
            cc_table[cc_table["class"] != "null"]['class_max'].max()
        x_df = np.where(x_df.notna() & ((x_df < minn) | (x_df >= maxx)), 0, x_df)
        if with_class_value:
            min_outlier_category = f"outlier_<{minn}"
            max_outlier_category = "outlier_>=inf" if np.isinf(maxx) else f"outlier_>={maxx}"
            c_df = pd.DataFrame(data=c_series.values, columns=["C_col"], index=c_series.index)
            c_df.loc[c_series.notna() & (c_series < minn), "C_col"] = min_outlier_category
            c_df.loc[c_series.notna() & (c_series >= maxx), "C_col"] = max_outlier_category
        return pd.Series(data=x_df), c_df

    def _fill_for_objects(self, x_df: pd.Series, c_series: pd.Series, c_df: pd.DataFrame, cc_table: pd.DataFrame,
                          with_class_value: bool) -> \
            Tuple[pd.Series, pd.DataFrame]:
        # unseed categories will remain zero
        res_df = pd.DataFrame(columns=["column", "transformed"])
        res_df["column"] = x_df
        res_df["transformed"] = 0
        categories = []
        for class_no in cc_table[cc_table["class"] != "null"]["class_no"]:
            res_df.loc[res_df["column"].isin(
                cc_table[cc_table['class_no'] == class_no]['class'].values[0].split(',')), "transformed"] = \
                cc_table[cc_table['class_no'] == class_no]['class_target_mean'].values[0] - \
                self.fit_output.target_mean
            if with_class_value:
                category = cc_table[cc_table['class_no'] == class_no]['class'].values[0]
                category_values = category.split(',')
                c_df.loc[c_df["C_col"].isin(category_values), "C_col"] = category
                categories.extend(category_values)
        res_df.loc[res_df["column"].isnull(), "transformed"] = \
            cc_table[cc_table['class'] == "null"]['class_target_mean'].values[0] - self.fit_output.target_mean
        if with_class_value:
            c_df.loc[c_series.notna() & ~c_series.isin(categories), "C_col"] = "Other (Unseen)"
            c_df.loc[c_series.isnull(), "C_col"] = "null"
        return res_df["transformed"], c_df

    def _fill_for_not_objects(self, x_df: pd.Series, c_series: pd.Series, c_df: pd.DataFrame, cc_table: pd.DataFrame,
                              with_class_value: bool) -> Tuple[pd.Series, pd.DataFrame]:
        for class_no in cc_table[cc_table["class"] != "null"]["class_no"]:
            minn = cc_table[cc_table["class_no"] == class_no]["class_min"].values[0]
            maxx = cc_table[cc_table["class_no"] == class_no]["class_max"].values[0]
            meann = cc_table[cc_table['class_no'] == class_no]['class_target_mean'].values[0]
            x_df = np.where((x_df >= minn) & (x_df < maxx), meann - self.fit_output.target_mean, x_df)
            if with_class_value:
                category = cc_table[cc_table["class_no"] == class_no]["class"].values[0]
                c_df.loc[((~c_series.isin([np.inf, -np.inf, np.nan])) & (c_series >= minn) & (
                        c_series < maxx)), "C_col"] = category
        return x_df, c_df



"""Includes NumericColumnStats class"""


class NumericColumnStats:
    """Numeric Column Stats"""

    def __init__(self, column_name: str, mean: float, std: float, percentile_25: float, percentile_50: float,
                 percentile_75: float):
        self.column_name = column_name
        self.mean = mean
        self.std = std
        self.percentile_25 = percentile_25
        self.percentile_50 = percentile_50
        self.percentile_75 = percentile_75

    def __eq__(self, other):
        if not isinstance(other, NumericColumnStats):
            return NotImplemented

        return self.mean == other.mean and self.std == other.std and self.percentile_25 == other.percentile_25 \
               and self.percentile_50 == other.percentile_50 and self.percentile_75 == other.percentile_75



"""Includes ScalingSettings class."""
from dataclasses import dataclass

from organon.ml.preprocessing.settings.enums.scaler_type import ScalerType


@dataclass
class ScalingSettings:
    """Settings for ScalingService."""
    strategy: ScalerType



"""Includes DqFileFullProcessInput class."""
from organon.idq.domain.settings.abstractions.dq_full_process_input import DqFullProcessInput
from organon.idq.domain.settings.calculation.dq_file_calculation_parameters import DqFileCalculationParameters
from organon.idq.domain.settings.comparison.dq_file_comparison_parameters import DqFileComparisonParameters


class DqFileFullProcessInput(DqFullProcessInput[DqFileCalculationParameters, DqFileComparisonParameters]):
    """DQ comparison parameters for text file source type"""
